import io
import os
import platform
import sys
import time
from abc import abstractmethod
from threading import RLock

from colorama import Fore, Style, Back

from aliyunpan.api.utils import str_of_size

__all__ = ['DATA', 'GLOBAL_VAR', 'Printer', 'Bar', 'FileBar', 'UploadBar', 'DownloadBar', 'HashBar']
os.system('')


class DATA(dict):
    def __init__(self, seq=None, **kwargs):
        if not seq:
            seq = {}
        super(DATA, self).__init__(seq, **kwargs)
        for key, value in seq.items():
            if isinstance(value, dict):
                self[key] = DATA(value)
            else:
                self[key] = value

    def to_dict(self):
        dict_ = {}
        for key, value in self.items():
            if isinstance(value, DATA):
                dict_[key] = value.to_dict()
            else:
                dict_[key] = value
        return dict_

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            return None

    def __setattr__(self, key, value):
        if isinstance(value, dict):
            self[key] = DATA(value)
        else:
            self[key] = value

    def __delattr__(self, item):
        del self[item]

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            super(DATA, self).__setitem__(key, DATA(value))
        else:
            super(DATA, self).__setitem__(key, value)


GLOBAL_VAR = DATA()
GLOBAL_VAR.tasks = {}
GLOBAL_VAR.file_set = set()


class Info:
    def __init__(self, info='', error=False, refresh_line=False, color=None):
        self.info = info
        self.error = error
        self.color = color
        if not self.color and self.error:
            self.color = Fore.RED
        self.refresh_line = refresh_line

    def __str__(self):
        info = self.info
        if self.color:
            info = Style.BRIGHT + Back.BLACK + self.color + info + Style.RESET_ALL
        if self.refresh_line:
            info = '\r' + info
        return info

    def __repr__(self):
        return self.__str__().__repr__()

    def __len__(self):
        return len(self.__str__())


class Flag:
    def __init__(self, status):
        self._start_flag = '*'
        self._success_flag = '+'
        self._fail_flag = '-'
        if status is None:
            self._flag = self._start_flag
            self.color = Fore.LIGHTYELLOW_EX
        elif status is True:
            self._flag = self._success_flag
            self.color = Fore.LIGHTGREEN_EX
        elif status is False:
            self._flag = self._fail_flag
            self.color = Fore.RED
        else:
            self._flag = status
            self.color = None

    def __str__(self):
        return self._flag


class OutPut(object):
    def __init__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        self._stdin = sys.stdin
        self._lock = RLock()

    @property
    @abstractmethod
    def output(self):
        pass

    @output.setter
    @abstractmethod
    def output(self, value):
        pass


class OutPutSingleton(OutPut):
    _instance = None
    _first_init = True

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if not self._first_init:
            return
        self._first_init = False
        super(OutPutSingleton, self).__init__()
        self._stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
        self._stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf-8')
        self._stdin = io.TextIOWrapper(sys.stdin.buffer, encoding='utf-8')
        self._print = self._output_gen()
        self._print.__next__()
        self.print_line = False

    def __del__(self):
        if platform.system() != 'Windows':
            self._stdout.write('\n')

    output = property(lambda self: self._print,
                      lambda self, value: (self._lock.acquire(), setattr(self, '_', None),
                                           self._print.send(value), self._lock.release())[1])

    def _output_gen(self):
        last_info = None
        while True:
            info = yield
            with self._lock:
                if self.print_line:
                    self._stdout.write('\n')
                    self.print_line = False
                if last_info:
                    if info.refresh_line:
                        if len(last_info) > len(info):
                            self._stdout.write('\r' + len(last_info) * ' ')
                if info.error:
                    self._stderr.write(str(info))
                else:
                    self._stdout.write(str(info))
                self._stderr.flush()
                self._stdout.flush()
            last_info = info


class Printer(OutPut):
    def __init__(self):
        super(Printer, self).__init__()
        self._info = '[{}]'
        self._existed_title = 'existed'
        self._upload_title = 'upload'
        self._rapid_upload_title = 'rapid'
        self._download_title = 'download'
        self._mkdir_title = 'mkdir'
        self._move_title = 'mv'
        self._link_info = ' -> '
        self._remove_title = 'rm'
        self._rename_title = 'rename'
        self._hash_title = 'hash'
        self._time_info = 'time:{:.2f}s'
        self.avg_info = 'avg:{:.2f}{}/s'
        self._error_info = 'Error:{}'
        self._aria2_title = 'aria2'
        self._mkdir_color = Fore.LIGHTCYAN_EX
        self._move_color = Fore.LIGHTMAGENTA_EX
        self._remove_color = Fore.LIGHTRED_EX
        self._rename_color = Fore.LIGHTBLUE_EX
        self._wait_color = Fore.MAGENTA
        self._error_color = Fore.RED
        self._print = OutPutSingleton()
        self._hash_size = 1024 ** 3
        self._output = True

    output = property(lambda self: self._print.output,
                      lambda self, value: (
                          self._lock.acquire(), setattr(self._print, 'output', value),
                          self._lock.release())[1] if self._output else None)

    def get_info(self, status, path, *args, existed=False, target_path=None, refresh_line=False):
        path_info = str(path) + self._link_info + str(target_path) if target_path else str(path)
        flag = Flag(status)
        info = self._info.format(flag)
        if existed:
            args = (*args[:1], self._existed_title, *args[1:])
        for i in args:
            if i:
                info += self._info.format(i)
        info = Info(info=info + path_info, error=status is False, refresh_line=refresh_line)
        info.color = flag.color
        return info

    def error_info(self, info, refresh_line=False):
        self.output = Info(self._error_info.format(info), color=self._error_color, error=True,
                           refresh_line=refresh_line)

    def wait_info(self, title=None, t=3, refresh_line=False):
        while t > 0:
            self.output = Info((title or '{time}秒后重试').format(time=t), color=self._wait_color,
                               refresh_line=refresh_line)
            t -= 1
            time.sleep(1)

    def upload_info(self, path, status=None, refresh_line=False, rapid_upload=False, t=None, average_speed=None, *args,
                    **kwargs):
        self._output = True
        t = self._time_info.format(t) if t else None
        average_speed = self.avg_info.format(*str_of_size(average_speed, tuple_=True)) if average_speed else None
        rapid_title = self._rapid_upload_title if rapid_upload else None
        info = self.get_info(status, path, self._upload_title, rapid_title, t, average_speed, *args,
                             refresh_line=refresh_line, **kwargs)
        self.output = info

    def download_info(self, path, status=None, refresh_line=False, t=None, average_speed=None, aria2=False, *args):
        self._output = True
        t = self._time_info.format(t) if t else None
        aria2 = self._aria2_title if aria2 else aria2
        average_speed = self.avg_info.format(*str_of_size(average_speed, tuple_=True)) if average_speed else None
        info = self.get_info(status, path, self._download_title, aria2, t, average_speed, *args,
                             refresh_line=refresh_line)
        self.output = info

    def mkdir_info(self, path, status=None, *args, **kwargs):
        self._output = True
        info = self.get_info(status, path, self._mkdir_title, *args, **kwargs)
        info.color = self._mkdir_color or info.color
        self.output = info

    def move_info(self, path, target_path, status=None, *args, **kwargs):
        self._output = True
        info = self.get_info(status, path, self._move_title, *args, target_path=target_path, **kwargs)
        info.color = self._move_color or info.color
        self.output = info

    def remove_info(self, path, status=None, *args, **kwargs):
        self._output = True
        info = self.get_info(status, path, self._remove_title, *args, **kwargs)
        info.error = False
        info.color = self._remove_color or info.color
        self.output = info

    def rename_info(self, path, name, status=None, *args, **kwargs):
        self._output = True
        info = self.get_info(status, path, self._rename_title, *args, target_path=name, **kwargs)
        info.color = self._rename_color or info.color
        self.output = info

    def hash_info(self, path, status=None, size=None, *args, **kwargs):
        info = self.get_info(status, path, self._hash_title, *args, **kwargs)
        self.output = info

    def refresh_line(self):
        if self._output:
            self._print.print_line = False
        self.output = Info(refresh_line=True)

    def print_line(self):
        if self._output:
            self._print.print_line = True

    def print_info(self, info, error=False, refresh_line=False, *args, **kwargs):
        self._output = True
        self.output = Info(info, error, refresh_line, *args, **kwargs)


class Bar(Printer):
    def __init__(self, title=None, refresh_interval=0.3):
        super(Bar, self).__init__()
        self._title = title or self.__class__.__name__
        self._upload_info = '{title}{:<3s} [{}{}] {:.2%} {:.2f}{unit}/s'
        self._ratio = 0.0
        self._start_time = time.time()
        self._time = time.time()
        self._average_speed = 0
        self._count = 0
        self._update_time = 0
        self._unit = 'it'
        self.refresh_interval = refresh_interval

    time = property(lambda self: time.time() - self._start_time)
    average_speed = property(lambda self: self._size / self.time if self.time else 0)

    def _get_average_speed(self, ratio, t):
        return (ratio - self._ratio) / t if t else 0

    def _format(self, *args, **kwargs):
        kwargs.setdefault('unit', self._unit)
        speed, unit = str_of_size(self._average_speed, tuple_=True)
        if not kwargs['unit']:
            kwargs['unit'] = unit
        return self._upload_info.format('.' * (4 - (self._count % 3 or 3)), '=' * int(self._ratio * 10),
                                        '*' * (10 - int(self._ratio * 10)), self._ratio,
                                        speed, title=self._title, *args, **kwargs)

    def update(self, ratio=None, refresh_line=False):
        self._count += 1
        t = 0
        if ratio is not None:
            t = time.time() - self._time
            self._time = time.time()
            average_speed = self._get_average_speed(ratio, t)
            if self._ratio != ratio and average_speed is not None:
                self._average_speed = average_speed
                self._ratio = ratio
        if t and time.time() - self._update_time >= self.refresh_interval:
            self._update_time = time.time()
            upload_info = self._format()
            self.output = Info(upload_info, refresh_line=refresh_line, color=Fore.LIGHTMAGENTA_EX)


class GetFileListBar(Bar):
    def __init__(self, max_depth, *args, **kwargs):
        super(GetFileListBar, self).__init__(*args, **kwargs)
        self._max_depth = max_depth
        self._depth = 0
        self._upload_info = '{title}{:<3s} [{}{}] {:.2%} [{depth}/{max_depth}] {:.2f}{unit}/s'
        self._title = 'get_file_list'
        self._time_out = 5
        self.refresh_interval = 0
        self._output = False

    def update(self, depth=0, *args, **kwargs):
        self._depth = depth
        if self.time > self._time_out:
            self._output = True
        super(GetFileListBar, self).update(*args, **kwargs)

    def _format(self, *args, **kwargs):
        return super(GetFileListBar, self)._format(depth=self._depth, max_depth=self._max_depth, *args, **kwargs)

    def _get_average_speed(self, ratio, t):
        if t:
            return super(GetFileListBar, self)._get_average_speed(ratio, t)


class FileBar(Bar):
    def __init__(self, title=None, size=0, *args, **kwargs):
        super(FileBar, self).__init__(*args, **kwargs)
        self._file_size = 1024 * 1024
        self._title = title or self.__class__.__name__
        self._size = size
        self._unit = None
        if self._size < self._file_size:
            self._output = False

    def _get_average_speed(self, ratio, t):
        return self._size * super(FileBar, self)._get_average_speed(ratio, t)


class UploadBar(FileBar):
    def __init__(self, size, *args, **kwargs):
        super(UploadBar, self).__init__(size=size, *args, **kwargs)
        self._title = self._upload_title


class DownloadBar(FileBar):
    def __init__(self, size, *args, **kwargs):
        super(DownloadBar, self).__init__(size=size, *args, **kwargs)
        self._title = self._download_title


class HashBar(FileBar):
    def __init__(self, size, *args, **kwargs):
        super(HashBar, self).__init__(size=size, *args, **kwargs)
        self._title = self._hash_title
        self._size = size
        self._time_out = 5
        if self._size < self._hash_size:
            self._output = False

    def update(self, *args, **kwargs):
        if self.time > self._time_out:
            self._output = True
        super(HashBar, self).update(*args, **kwargs)

    def hash_info(self, *args, **kwargs):
        super(HashBar, self).hash_info(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import click
from click_aliases import ClickAliasedGroup

from aliyunpan.about import __version__
from aliyunpan.api.utils import logger
from aliyunpan.cli.cli import Commander


@click.group(cls=ClickAliasedGroup)
@click.help_option('-h', '--help')
@click.version_option(version=__version__)
@click.option('-c', '--config-file', type=click.Path(), help='Specify the configuration file.',
              default='~/.config/aliyunpan.yaml', show_default=True)
@click.option('-t', '--refresh-token', type=click.STRING, help='Specify REFRESH_TOKEN.')
@click.option('-u', '--username', type=click.STRING, help='Specify USERNAME.')
@click.option('-p', '--password', type=click.STRING, help='Specify PASSWORD.')
@click.option('-d', '--depth', type=click.INT, help='File recursion depth.', default=3, show_default=True)
@click.option('-D', '--debug', is_flag=True, help='Debug mode.')
@click.option('-T', '--timeout', type=click.FLOAT, help='Api request timeout.')
@click.option('-id', '--drive-id', type=click.STRING, help='Specify DRIVE_ID.')
@click.option('-a', '--album', is_flag=True, help='Specify album')
def cli(config_file, refresh_token, username, password, depth, debug, timeout, drive_id, album):
    logger.info(f'Version:{__version__}')
    if debug:
        logger.setLevel('DEBUG')
    commander.init(config_file=None if refresh_token or username else config_file,
                   refresh_token=None if username else refresh_token, username=username, password=password, depth=depth,
                   timeout=timeout, drive_id=drive_id, album=album)


@cli.command(aliases=['l', 'list', 'dir'], help='List files.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='root')
@click.option('-l', is_flag=True, help='View details.')
def ls(path, l):
    commander.ls(path, l)


@cli.command(aliases=['search'], help='Search for file.')
@click.help_option('-h', '--help')
@click.argument('query', type=click.STRING)
@click.option('-l', is_flag=True, help='View details.')
def search(query, l):
    commander.ls(path=None, l=l, query=query)


@cli.command(aliases=['delete', 'del'], help='Delete Files.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path())
def rm(path):
    commander.rm(path)


@cli.command(aliases=['r'], help='Rename file.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path())
@click.argument('name', type=click.STRING)
def rename(path, name):
    commander.rename(path, name)


@cli.command(aliases=['move'], help='Move files.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path())
@click.argument('target_path', type=click.Path())
def mv(path, target_path):
    commander.mv(path, target_path)


@cli.command(aliases=['u'], help='Upload files.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='')
@click.option('-p', '--file', multiple=True, help='Select multiple files.', type=click.Path())
@click.argument('upload_path', default='root')
@click.option('-t', '--time-out', type=click.FLOAT, help='Chunk upload timeout(sec).', default=10.0, show_default=True)
@click.option('-r', '--retry', type=click.INT, help='number of retries.', default=3, show_default=True)
@click.option('-f', '--force', is_flag=True, help='Force overlay file.')
@click.option('-s', '--share', is_flag=True, help='Specify the shared sequence file.')
@click.option('-cs', '--chunk-size', type=click.INT, help='Chunk size(byte).')
@click.option('-c', is_flag=True, help='Breakpoint continuation.')
def upload(path, file, upload_path, time_out, retry, force, share, chunk_size, c):
    if not path and not file:
        raise click.MissingParameter(param=click.get_current_context().command.params[2])
    else:
        path_list = {*file, path}
    commander.upload(path_list, upload_path, time_out, retry, force, share, chunk_size, c)


@cli.command(aliases=['m'], help='Create folder.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='')
def mkdir(path):
    commander.mkdir(path)


@cli.command(aliases=['d'], help='Download files.',
             context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='')
@click.option('-p', '--file', multiple=True, help='Select multiple files.', type=click.Path())
@click.argument('save_path', type=click.Path(), default='')
@click.option('-s', 'share', is_flag=True, help='Specify the shared sequence file')
@click.option('-cs', '--chunk-size', type=click.INT, help='Chunk size(byte).', default=1048576, show_default=True)
@click.option('-a', '--aria2', is_flag=True, help='Send to aria2.')
@click.pass_context
def download(ctx, path, file, save_path, share, chunk_size, aria2):
    if not path and not file:
        raise click.MissingParameter(param=click.get_current_context().command.params[2])
    else:
        file_list = {*file, path}
    kwargs = {}
    for i in ctx.args:
        kwargs[i.split('=')[0]] = i.split('=')[1]
    commander.download(file_list, save_path=save_path, share=share, chunk_size=chunk_size, aria2=aria2, **kwargs)


@cli.command(aliases=['t', 'show'], help='View file tree.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='root')
def tree(path):
    commander.tree(path)


@cli.command(aliases=['s'], help='Share file sharing link.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='')
@click.option('-p', '--file', multiple=True, help='Select multiple files.', type=click.Path())
@click.option('-f', '--file-id', multiple=True, type=click.STRING, help='File id.')
@click.option('-t', '--expire-sec', type=click.INT, help='Link expiration time(Max 14400 or permanent).')
@click.option('-l', '--share-link', is_flag=True, help='Output share link.', default=True, show_default=True)
@click.option('-d', '--download-link', is_flag=True, help='Output download link.', show_default=True)
@click.option('-s', '--save', is_flag=True, help='Save to cloud disk and local.', show_default=True)
@click.option('-S', '--share-official', is_flag=True, help='Output official share link.', show_default=True)
def share(path, file, file_id, expire_sec, share_link, download_link, save, share_official):
    if not path and not file and not file_id:
        raise click.MissingParameter(param=click.get_current_context().command.params[2])
    else:
        path_list = list(filter(None, {*file, path}))
        file_id_list = list(filter(None, {*file_id}))
    if share_official:
        commander.share_link(path_list, file_id_list, expire_sec)
    elif path_list:
        for path in path_list:
            if path:
                commander.share(path, expire_sec, share_link, download_link, save)
    else:
        raise click.MissingParameter(param=click.get_current_context().command.params[1])


@cli.command(aliases=['c'], help='Show file content.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path(), default='')
@click.option('-e', '--encoding', type=click.STRING, default='utf-8', show_default=True)
def cat(path, encoding):
    click.echo(commander.cat(path, encoding))


@cli.command(aliases=['sync'], help='Synchronize files.')
@click.help_option('-h', '--help')
@click.argument('path', type=click.Path())
@click.argument('upload_path', default='root')
@click.option('-cs', '--chunk-size', type=click.INT, help='Chunk size(byte).')
@click.option('-t', '--time-out', type=click.FLOAT, help='Chunk upload timeout(sec).', default=10.0, show_default=True)
@click.option('-r', '--retry', type=click.INT, help='number of retries.', default=3, show_default=True)
@click.option('--sync-time', type=click.FLOAT, help='Synchronization interval time(sec).')
def sync(path, upload_path, time_out, chunk_size, retry, sync_time):
    commander.sync(path, upload_path, sync_time, time_out, chunk_size, retry)


@cli.command(aliases=['tui'], help='Text-based User Interface.')
@click.help_option('-h', '--help')
def tui():
    commander.tui()


@cli.command(aliases=['r', 'refresh_token'], help='Get refresh_token.')
@click.help_option('-h', '--help')
@click.option('--refresh', '-r', is_flag=True, help='Refresh the token of the configuration file.')
@click.option('--refresh-time', '-t', type=click.FLOAT, help='Auto refresh token interval time(sec).')
@click.option('--change', '-c', type=click.STRING, help='Set new refresh_token.')
def token(refresh, refresh_time, change):
    if refresh:
        commander.disk.token_refresh()
    elif refresh_time:
        commander.auto_refresh_token(refresh_time)
    elif change:
        if commander.disk.token_refresh(change):
            commander.disk.refresh_token = change
    click.echo(commander.disk.refresh_token)


commander: Commander


def main():
    global commander
    commander = Commander(init=False)
    cli()


if __name__ == '__main__':
    main()



class AliyunpanException(BaseException):
    def __init__(self, message='', *args):
        super(AliyunpanException, self).__init__(*args)
        self.message = message

    def __str__(self):
        return self.message


class InvalidToken(AliyunpanException):
    """无效的Token"""


class InvalidRefreshToken(InvalidToken):
    """无效的refresh_token"""

    def __str__(self):
        return self.message or 'Is not a valid refresh_token.'


class InvalidAccessToken(InvalidToken):
    """无效的access_token"""

    def __str__(self):
        return self.message or 'Is not a valid access_token.'


class InvalidPassword(AliyunpanException):
    """无效的密码"""


class LoginFailed(AliyunpanException):
    """登录失败"""


class ConfigurationFileError(AliyunpanException, FileNotFoundError):
    """配置文件错误"""


class ConfigurationFileNotFoundError(ConfigurationFileError):
    """找不到配置文件"""


class InvalidConfiguration(ConfigurationFileError):
    """无效的配置文件"""


class InvalidParameter(AliyunpanException):
    """参数错误"""


class InvalidContentHash(InvalidParameter):
    """Hash不匹配"""


class InvalidExpiration(InvalidParameter):
    """错误的Expiration"""


class RequestExpired(AliyunpanException):
    """请求过期"""


class UploadUrlExpired(RequestExpired):
    """上传链接过期"""


class UploadUrlFailedRefresh(UploadUrlExpired):
    """上传链接刷新失败"""


class BadResponseCode(AliyunpanException):
    """错误的响应代码"""


class InvalidPartNumber(InvalidParameter):
    """分块数量错误"""


class PartNumberOverLimit(InvalidPartNumber):
    """分块上限"""

    def __str__(self):
        return self.message or '\nPart Number must be an integer between 1 and 10000, inclusive.\n' \
                               'Please increase the size of chunk_size.'


class PartNotSequential(AliyunpanException):
    """上传序列错误"""

    def __str__(self):
        return self.message or 'For sequential multipart upload,' \
                               'you must upload or complete parts with sequential part number.'


class FileShareNotAllowed(AliyunpanException):
    """文件无法分享"""


class AliyunpanCode(object):
    existed = 'AlreadyExist.File'
    token_invalid = 'AccessTokenInvalid'
    invalid_content_hash = 'InvalidParameter.ContentHash'
    not_found_file = 'NotFound.File'
    request_expired = 403
    part_already_exist = 409
    part_not_sequential = 400
    Forbidden = 'Forbidden'
    InvalidExpiration = 'InvalidParameter.Expiration'
    FileShareNotAllowed = 'FileShareNotAllowed'



import json
import sys
from threading import RLock

import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from aliyunpan.api.utils import logger
from aliyunpan.common import *
from aliyunpan.exceptions import AliyunpanCode, InvalidAccessToken, InvalidRefreshToken, LoginFailed

__all__ = ['Req']

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class Req:
    _instance = None
    _first_init = True

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, disk=None):
        if not self._first_init:
            return
        self._first_init = False
        self._disk = disk
        self._lock = RLock()
        self._retry_num = 3
        self._session = requests.Session()
        self._timeout = 5
        self._verify = False
        self._host_url = 'https://www.aliyundrive.com/'
        self._headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                          'Chrome/87.0.4280.88 Safari/537.36',
            'Referer': self._host_url
        }

    timeout = property(lambda self: self._timeout, lambda self, value: setattr(self, '_timeout', value))
    verify = property(lambda self: self._verify, lambda self, value: setattr(self, '_verify', value))
    retry_num = property(lambda self: self._retry_num, lambda self, value: setattr(self, '_retry_num', value))

    def _req(self, method, *args, **kwargs):
        try:
            kwargs.setdefault('timeout', self._timeout)
            kwargs.setdefault('verify', self._verify)
            kwargs.setdefault('stream', False)
            stream = kwargs['stream']
            kwargs.setdefault('depth', self._retry_num)
            depth = kwargs.pop('depth')
            kwargs['headers'] = kwargs['headers'] if 'headers' in kwargs else {}
            kwargs['headers'].update(self._headers)
            if 'access_token' in kwargs:
                if kwargs['access_token']:
                    kwargs['headers']['Authorization'] = kwargs['access_token']
                else:
                    kwargs['headers']['Authorization'] = None
                del kwargs['access_token']
            else:
                kwargs['headers']['Authorization'] = self._disk.access_token if self._disk else GLOBAL_VAR.access_token
            r = getattr(self._session, method.lower())(*args, **kwargs)
            try:
                logger.debug(r.status_code)
                if depth:
                    self._lock.acquire()
                    if self._disk and not stream and json.loads(r.text)['code'] == AliyunpanCode.token_invalid \
                            or self._disk.refresh_token_expires_sec < 600:
                        depth -= 1
                        try:
                            self._disk.access_token = self._disk.get_access_token()
                        except InvalidRefreshToken:
                            self._disk.login()
                        self._lock.release()
                        return self._req(method, depth=depth, *args, **kwargs)
                    self._lock.release()
                else:
                    raise InvalidAccessToken
            except (KeyboardInterrupt, InvalidAccessToken, LoginFailed):
                raise
            except json.decoder.JSONDecodeError:
                logger.debug(r.text)
            except KeyError:
                pass
            except:
                logger.debug(sys.exc_info())
            finally:
                try:
                    self._lock.release()
                except RuntimeError:
                    pass
            return r
        except requests.exceptions.RequestException:
            raise

    def get(self, *args, **kwargs) -> requests.models.Response:
        return self._req('get', *args, **kwargs)

    def post(self, *args, **kwargs) -> requests.models.Response:
        return self._req('post', *args, **kwargs)

    def put(self, *args, **kwargs) -> requests.models.Response:
        return self._req('put', *args, **kwargs)

    def head(self, *args, **kwargs) -> requests.models.Response:
        return self._req('head', *args, **kwargs)

    def options(self, *args, **kwargs) -> requests.models.Response:
        return self._req('options', *args, **kwargs)

    def req(self, method, *args, **kwargs) -> requests.models.Response:
        return self._req(method, *args, **kwargs)



import sys
import time
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from threading import RLock

import requests

from aliyunpan.api import ua
from aliyunpan.api.req import *
from aliyunpan.api.type import UserInfo, AlibumInfo
from aliyunpan.api.utils import *
from aliyunpan.common import *
from aliyunpan.exceptions import InvalidRefreshToken, AliyunpanException, AliyunpanCode, LoginFailed, \
    InvalidContentHash, UploadUrlExpired, UploadUrlFailedRefresh, PartNumberOverLimit, BadResponseCode, \
    PartNotSequential, InvalidExpiration, FileShareNotAllowed

__all__ = ['AliyunPan']


class AliyunPan(object):

    def __init__(self, refresh_token: str = None, album: bool = False):
        self._req = Req(self)
        self._user_info = None
        self._alibum_info = None
        self._album = album
        self._access_token = None
        self._drive_id = None
        self._username = None
        self._password = None
        self._refresh_token = refresh_token
        self._refresh_token_expires = None
        self._access_token_gen_ = self._access_token_gen()
        self._drive_id_gen_ = self._drive_id_gen()
        self._chunk_size = 524288
        self._print = Printer()
        self._lock = RLock()

    refresh_token = property(lambda self: self._refresh_token,
                             lambda self, value: setattr(self, '_refresh_token', value))
    access_token = property(lambda self: (self._lock.acquire(), next(self._access_token_gen_), self._lock.release())[1],
                            lambda self, value: setattr(self, '_access_token', value))
    drive_id = property(lambda self: (self._lock.acquire(), next(self._drive_id_gen_), self._lock.release())[1],
                        lambda self, value: setattr(self, '_drive_id', value))
    album = property(lambda self: self._album, lambda self, value: setattr(self, '_album', value))

    def login(self, username: str = None, password: str = None):
        """
        登录api
        https://github.com/zhjc1124/aliyundrive
        :param username:
        :param password:
        :return:
        """
        if username:
            self._username = username
        else:
            username = self._username
        if password:
            self._password = password
        else:
            password = self._password
        if not username and not password:
            raise LoginFailed
        password2 = encrypt(password)
        LOGIN = {
            'method': 'POST',
            'url': 'https://passport.aliyundrive.com/newlogin/login.do',
            'data': {
                'loginId': username,
                'password2': password2,
                'appName': 'aliyun_drive',
                'ua': ua.get_ua()
            }
        }
        logger.info('Logging in.')
        r = self._req.req(**LOGIN, access_token=False)
        logger.debug(r.json())
        try:
            data = parse_biz_ext(r.json()['content']['data']['bizExt'])
            logger.debug(data)
            access_token = data['pds_login_result']['accessToken']
            self._access_token = access_token
            GLOBAL_VAR.access_token = access_token
            refresh_token = data['pds_login_result']['refreshToken']
            self._refresh_token = refresh_token
            GLOBAL_VAR.refresh_token = refresh_token
            drive_id = data['pds_login_result']['defaultDriveId']
            self._drive_id = drive_id
            GLOBAL_VAR.drive_id = drive_id
            return self._refresh_token
        except KeyError:
            pass
        raise LoginFailed

    def get_file_list(self, parent_file_id: str = 'root', next_marker: str = None) -> list:
        """
        获取文件列表
        :param parent_file_id:
        :param next_marker:
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/file/list'
        json = {"drive_id": self.drive_id, "parent_file_id": parent_file_id, 'fields': '*', 'marker': next_marker}
        logger.info(f'Get the list of parent_file_id {parent_file_id}.')
        r = self._req.post(url, json=json)
        logger.debug(r.json())
        if 'items' not in r.json():
            return []
        file_list = r.json()['items']
        if 'next_marker' in r.json() and r.json()['next_marker'] and next_marker != r.json()['next_marker']:
            file_list.extend(self.get_file_list(parent_file_id, r.json()['next_marker']))
        return file_list

    def delete_file(self, file_id: str):
        """
        删除文件
        :param file_id:
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/batch'
        json = {'requests': [{'body': {'drive_id': self.drive_id, 'file_id': file_id},
                              'headers': {'Content-Type': 'application/json'},
                              'id': file_id, 'method': 'POST',
                              'url': '/recyclebin/trash'}], 'resource': 'file'}
        logger.info(f'Delete file {file_id}.')
        r = self._req.post(url, json=json)
        logger.debug(r.text)
        if r.status_code == 200:
            return r.json()['responses'][0]['id']
        return False

    def move_file(self, file_id: str, parent_file_id: str):
        """
        移动文件
        :param file_id:
        :param parent_file_id:
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/batch'
        json = {"requests": [{"body": {"drive_id": self.drive_id, "file_id": file_id,
                                       "to_parent_file_id": parent_file_id},
                              "headers": {"Content-Type": "application/json"},
                              "id": file_id, "method": "POST", "url": "/file/move"}],
                "resource": "file"}
        logger.info(f'Move files {file_id} to {parent_file_id}')
        r = self._req.post(url, json=json)
        if r.status_code == 200:
            if 'message' in r.json()['responses'][0]['body']:
                print(r.json()['responses'][0]['body']['message'])
                return False
            return r.json()['responses'][0]['id']
        return False

    def update_file(self, file_id: str, name: str):
        """
        重命名文件
        :param file_id:
        :param name:
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/file/update'
        json = {"drive_id": self.drive_id, "file_id": file_id, "name": name,
                "check_name_mode": "refuse"}
        r = self._req.post(url, json=json)
        logger.info(f'Rename {file_id} to {name}')
        logger.debug(r.json())
        if r.status_code == 200:
            return r.json()['file_id']
        return r.json()['code']

    def get_user_info(self) -> UserInfo:
        """
        获取用户信息
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/user/get'
        logger.info('Get user information.')
        r = self._req.post(url, json={})
        user_info = r.json()
        id_ = user_info['user_id']
        nick_name = user_info['nick_name']
        ctime = user_info['created_at']
        phone = user_info['phone']
        drive_id = user_info['default_drive_id']
        user_info = UserInfo(id=id_, nick_name=nick_name, ctime=ctime, phone=phone, drive_id=drive_id)
        logger.debug(user_info)
        self._user_info = user_info
        GLOBAL_VAR.user_info = user_info
        return user_info

    def get_albums_info(self) -> AlibumInfo:
        url = 'https://api.aliyundrive.com/adrive/v1/user/albums_info'
        r = self._req.post(url, json={})
        data = r.json()['data']
        drive_id = data['driveId']
        drive_name = data['driveName']
        alibum_info = AlibumInfo(drive_name=drive_name, drive_id=drive_id)
        self._alibum_info = alibum_info
        return alibum_info

    def create_file(self, file_name: str, parent_file_id: str = 'root', file_type: bool = False,
                    json: dict = None, force: bool = False) -> requests.models.Response:
        """
        创建文件
        :param file_name:
        :param parent_file_id:
        :param file_type:
        :param json:
        :param force:
        :return:
        """
        j = {
            "name": file_name,
            "drive_id": self.drive_id,
            "parent_file_id": parent_file_id,
            "content_hash_name": "sha1",
        }
        if file_type:
            f_type = 'file'
            check_name_mode = 'auto_rename'
            if force:
                check_name_mode = 'refuse'
        else:
            f_type = 'folder'
            check_name_mode = 'refuse'
        j.update({"type": f_type, 'check_name_mode': check_name_mode})
        if json:
            j.update(json)
        # 申请创建文件
        # url = 'https://api.aliyundrive.com/v2/file/create'
        # if 'proof_code' in j:
        url = 'https://api.aliyundrive.com/adrive/v2/file/createWithFolders'
        logger.info(f'Create file {file_name} in file {parent_file_id}.')
        retry_num = 3
        while retry_num:
            try:
                r = self._req.post(url, json=j)
            except requests.exceptions.RequestException:
                logger.error(sys.exc_info())
                retry_num -= 1
            else:
                break
        else:
            raise requests.exceptions.RequestException
        logger.debug(j)
        if force and 'exist' in r.json():
            self.delete_file(r.json()['file_id'])
            return self.create_file(file_name, parent_file_id, file_type, json)
        return r

    def upload_file(self, parent_file_id: str = 'root', path: str = None, upload_timeout: float = 10,
                    retry_num: int = 3, force: bool = False, chunk_size: int = None, c: bool = False):
        """
        上传文件
        :param parent_file_id: 上传目录的id
        :param path: 上传文件路径
        :param upload_timeout: 分块上传超时时间
        :param retry_num:
        :param force: 强制覆盖
        :param chunk_size: 分块大小
        :param c: 断点续传
        :return:
        """
        path = Path(path)
        file_size = path.stat().st_size
        file_name = path.name
        self._chunk_size = chunk_size or self._chunk_size
        # 获取sha1
        content_hash = get_sha1(path, self._chunk_size)
        while True:
            # 分片列表
            part_info_list = []
            count = int(file_size / self._chunk_size) + 1
            for i in range(count):
                part_info_list.append({"part_number": i + 1})
            if len(part_info_list) > 10000:
                if chunk_size:
                    raise PartNumberOverLimit
                self._chunk_size = int(file_size / 1000)
                continue
            break
        proof_code = get_proof_code(get_file_byte(path, self.access_token))
        json = {"size": file_size, "part_info_list": part_info_list, "content_hash": content_hash,
                'proof_code': proof_code, 'proof_version': 'v1'}
        path_list = []
        existed = False
        # 已存在任务
        if content_hash in GLOBAL_VAR.tasks:
            # 是否是列表或可迭代对象
            if isinstance(GLOBAL_VAR.tasks[content_hash].path, Iterable) and not isinstance(
                    GLOBAL_VAR.tasks[content_hash].path, str):
                path_list.extend(GLOBAL_VAR.tasks[content_hash].path)
            else:
                path_list.append(GLOBAL_VAR.tasks[content_hash].path)
            path_list = list(set([str(get_real_path(path_)) for path_ in path_list]))
            flag = False
            # 云盘是否存在该文件
            if GLOBAL_VAR.tasks[content_hash].upload_time and path_list:
                existed = True
                for path_ in path_list:
                    # 是否存在路径
                    if get_real_path(path) == get_real_path(path_):
                        flag = True
                        break
            # 云盘存在该文件且存在路径且断点续传
            if existed and flag and c:
                # 已存在跳过上传
                self._print.upload_info(path, status=True, existed=True)
                self._print.print_line()
                GLOBAL_VAR.tasks[content_hash].path = path_list[0] if len(path_list) == 1 else path_list
                GLOBAL_VAR.file_set.add((content_hash, str(get_real_path(path))))
                return GLOBAL_VAR.tasks[content_hash].file_id
        # 断点续传且已存在任务且云盘不存在该文件
        if c and content_hash in GLOBAL_VAR.tasks and not existed:
            upload_id = GLOBAL_VAR.tasks[content_hash].upload_id
            file_id = GLOBAL_VAR.tasks[content_hash].file_id
            self._chunk_size = GLOBAL_VAR.tasks[content_hash].chunk_size
            part_number = GLOBAL_VAR.tasks[content_hash].part_number
            try:
                # 获取上传链接列表
                part_info_list = self.get_upload_url(path, upload_id, file_id, self._chunk_size, part_number)
                if not part_info_list:
                    # 重新上传
                    if str(get_real_path(path)) in path_list:
                        # 删除未上传成功的任务
                        del path_list[str(get_real_path(path))]
                    GLOBAL_VAR.tasks[content_hash].path = path_list[0] if len(path_list) == 1 else path_list
                    return self.upload_file(parent_file_id=parent_file_id, path=path, upload_timeout=upload_timeout,
                                            retry_num=retry_num, force=force, chunk_size=self._chunk_size, c=c)
            except FileExistsError:
                # 漏网之鱼
                self._print.upload_info(path, status=True, existed=True)
                self._print.print_line()
                path_list.append(str(get_real_path(path)))
                path_list = list(set(path_list))
                GLOBAL_VAR.tasks[content_hash].path = path_list[0] if len(path_list) == 1 else path_list
                GLOBAL_VAR.file_set.add((content_hash, str(get_real_path(path))))
                return GLOBAL_VAR.tasks[content_hash].file_id
        else:
            # 申请创建文件
            r = self.create_file(file_name=file_name, parent_file_id=parent_file_id, file_type=True, json=json,
                                 force=force)
            if 'rapid_upload' not in r.json():
                message = r.json()['message']
                logger.error(message)
                raise AliyunpanException(message)
            task_info = {'path': str(get_real_path(path)), 'upload_id': None,
                         'file_id': None, 'chunk_size': self._chunk_size,
                         'part_number': None}
            rapid_upload = r.json()['rapid_upload']
            # 快速上传成功
            if rapid_upload:
                self._print.upload_info(path, status=True, rapid_upload=True)
                self._print.print_line()
                file_id = r.json()['file_id']
                task_info['file_id'] = file_id
                task_info['upload_time'] = time.time()
                GLOBAL_VAR.tasks[content_hash] = task_info
                GLOBAL_VAR.file_set.add((content_hash, str(get_real_path(path))))
                if existed:
                    # 同hash不同路径
                    path_list.append(str(get_real_path(path)))
                    path_list = list(set(path_list))
                    GLOBAL_VAR.tasks[content_hash].path = path_list[0] if len(path_list) == 1 else path_list
                return file_id
            else:
                upload_id = r.json()['upload_id']
                file_id = r.json()['file_id']
                part_info_list = r.json()['part_info_list']
                task_info['upload_id'] = upload_id
                task_info['file_id'] = file_id
                task_info['part_number'] = 1
                GLOBAL_VAR.tasks[content_hash] = task_info
        upload_bar = UploadBar(size=file_size)
        upload_bar.upload_info(path)
        upload_bar.print_line()
        upload_bar.update(refresh_line=False)
        logger.debug(f'upload_id: {upload_id}, file_id: {file_id}, part_info_list: {part_info_list}')
        part_info_list = Iter(part_info_list)
        for i in part_info_list:
            part_number, upload_url = i['part_number'], i['upload_url']
            if not upload_url:
                continue
            GLOBAL_VAR.tasks[content_hash].part_number = part_number
            # 分块读取
            with path.open('rb') as f:
                f.seek((part_number - 1) * self._chunk_size)
                chunk = f.read(self._chunk_size)
            if not chunk:
                break
            size = len(chunk)
            retry_count = 0
            while True:
                upload_bar.update(refresh_line=True)
                logger.debug(
                    f'(upload_id={upload_id}, file_id={file_id}, size={size}): Upload part of {part_number} to {upload_url}.')
                try:
                    # 开始上传
                    r = self._req.put(upload_url, data=chunk, timeout=upload_timeout, access_token=False)
                    if r.status_code == AliyunpanCode.request_expired:
                        raise UploadUrlExpired
                    elif r.status_code == AliyunpanCode.part_already_exist:
                        pass
                    elif r.status_code == AliyunpanCode.part_not_sequential:
                        raise PartNotSequential
                    elif r.status_code != 200:
                        logger.error(r.status_code)
                        raise BadResponseCode
                    break
                except (requests.exceptions.SSLError, requests.exceptions.ConnectionError,
                        requests.exceptions.ReadTimeout):
                    logger.warn('Upload timeout.')
                    if retry_count is retry_num:
                        self._print.error_info(f'上传超时{retry_num}次，即将重新上传', refresh_line=True)
                        time.sleep(1)
                        return self.upload_file(parent_file_id=parent_file_id, path=path, upload_timeout=upload_timeout,
                                                retry_num=retry_num, force=force, chunk_size=self._chunk_size, c=c)
                    self._print.error_info('上传超时', refresh_line=True)
                    retry_count += 1
                    time.sleep(1)
                except KeyboardInterrupt:
                    raise
                except UploadUrlExpired:
                    info = f'Part {part_number} upload request has expired.'
                    logger.warning(info)
                    self._print.error_info(info, refresh_line=True)
                    time.sleep(1)
                    part_info_list_ = self.get_upload_url(path=path, upload_id=upload_id, file_id=file_id,
                                                          chunk_size=self._chunk_size)
                    if part_info_list_:
                        part_info_list.iter = part_info_list_
                        part_info = [i for i in part_info_list if i['part_number'] == part_number][0]
                        upload_url = part_info['upload_url']
                        logger.info(f'The upload_url of Part {part_number} has been refreshed.')
                        logger.debug(upload_url)
                    else:
                        logger.error(f'The upload_url of Part {part_number} failed to refresh.')
                        raise UploadUrlFailedRefresh
                except (BadResponseCode, PartNotSequential):
                    raise
                except:
                    logger.error(sys.exc_info())
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    self._print.error_info(exc_type.__name__, refresh_line=True)
                    time.sleep(1)
                self._print.wait_info(refresh_line=True)
            k = part_number / len(part_info_list)
            upload_bar.update(ratio=k, refresh_line=True)
        file_info = None
        try:
            file_info = self.complete(file_id, upload_id)
        except InvalidContentHash:
            upload_bar.upload_info(path, status=False, refresh_line=True)
            if get_real_path(log_file) != get_real_path(path):
                raise
        if file_info:
            upload_bar.upload_info(path, status=True, t=upload_bar.time, average_speed=upload_bar.average_speed,
                                   refresh_line=True)
            self._print.print_line()
            GLOBAL_VAR.tasks[content_hash].upload_time = time.time()
            GLOBAL_VAR.file_set.add((content_hash, str(get_real_path(path))))
            return file_info
        else:
            upload_bar.upload_info(path, status=False, refresh_line=True)
            self._print.print_line()
            return False

    def complete(self, file_id, upload_id):
        """
        上传成功保存文件
        """
        url = 'https://api.aliyundrive.com/v2/file/complete'
        json = {
            "ignoreError": True,
            "drive_id": self.drive_id,
            "file_id": file_id,
            "upload_id": upload_id
        }
        r = self._req.post(url, json=json)
        logger.debug(r.json())
        if 'code' in r.json() and r.json()['code'] == AliyunpanCode.invalid_content_hash:
            raise InvalidContentHash
        if r.status_code == 200:
            return r.json()
        return False

    def get_upload_url(self, path: str, upload_id: str, file_id: str, chunk_size: int, part_number: int = 1) -> list:
        """
        获取上传地址
        :param path:
        :param upload_id:
        :param file_id:
        :param chunk_size:
        :param part_number:
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/file/get_upload_url'
        path = Path(path)
        file_size = path.stat().st_size
        part_info_list = []
        count = int(file_size / chunk_size) + 1
        for i in range(count):
            part_info_list.append({"part_number": i + 1})
        json = {
            "drive_id": self.drive_id,
            "file_id": file_id,
            "part_info_list": part_info_list,
            "upload_id": upload_id,
        }
        r = self._req.post(url, json=json)
        if 'code' in r.json() and r.json()['code'] == AliyunpanCode.existed:
            raise FileExistsError
        part_info_list = r.json()['part_info_list']
        for i in part_info_list[:part_number - 1]:
            i['upload_url'] = ''
        return part_info_list

    def token_refresh(self, refresh_token: str = None):
        # url = 'https://websv.aliyundrive.com/token/refresh'
        # json = {"refresh_token": self.refresh_token}
        url = 'https://auth.aliyundrive.com/v2/account/token'
        json = {"refresh_token": refresh_token or self.refresh_token, 'grant_type': 'refresh_token'}
        logger.info(f'Token has been refreshed.')
        r = self._req.post(url, json=json, access_token=False)
        try:
            self._refresh_token = r.json()['refresh_token']
            self._refresh_token_expires = time.time() + r.json()['expires_in']
        except KeyError:
            raise InvalidRefreshToken
        return r.json()

    @property
    def refresh_token_expires_sec(self):
        return self._refresh_token_expires - time.time()

    def get_access_token(self) -> str:
        """
        获取access_token
        :return:
        """
        token_refresh_data = self.token_refresh()
        access_token = token_refresh_data['access_token']
        GLOBAL_VAR.refresh_token = self.refresh_token
        GLOBAL_VAR.access_token = access_token
        logger.debug(access_token)
        return access_token

    def _access_token_gen(self) -> str:
        """
        access_token生成器
        :return:
        """
        access_token = None
        while True:
            if self._access_token:
                yield self._access_token
            elif access_token:
                yield access_token
            else:
                access_token = self.get_access_token()

    def get_drive_id(self) -> str:
        """
        获取drive_id
        :return:
        """
        if self._album:
            drive_id = self.get_albums_info().drive_id
        else:
            drive_id = self.get_user_info().drive_id
        self._drive_id = drive_id
        GLOBAL_VAR.drive_id = drive_id
        return drive_id

    def _drive_id_gen(self) -> str:
        """
        drive_id生成器
        :return:
        """
        drive_id = None
        while True:
            if self._drive_id:
                yield self._drive_id
            elif drive_id:
                yield drive_id
            else:
                drive_id = self.get_drive_id()

    def get_download_url(self, file_id, expire_sec=14400, category=None) -> str:
        """
        获取分享链接
        :param file_id:
        :param expire_sec: 文件过期时间（秒）
        :param category:
        :return:
        """
        url = 'https://api.aliyundrive.com/v2/file/get_download_url'
        illegal_url = 'https://pds-system-file.oss-cn-beijing.aliyuncs.com/illegal.mp4'
        json = {'drive_id': self.drive_id, 'file_id': file_id, 'expire_sec': expire_sec}
        logger.info(f'Get file {file_id} download link, expiration time {expire_sec} seconds.')
        r = self._req.post(url, json=json)
        url = r.json()['url'] if 'url' in r.json() else ''
        if not url or url == illegal_url:
            url_dict = self.get_play_info(file_id, expire_sec, category) if category else \
                self.get_play_info(file_id, expire_sec, 'video') or self.get_play_info(file_id, expire_sec, 'audio')
            if url_dict:
                url = list(url_dict.values())[-1]
            elif 'internal_url' in r.json() and r.json()['internal_url']:
                url = r.json()['internal_url']
        logger.debug(f'file_id:{file_id},expire_sec:{expire_sec},url:{url}')
        return url

    def save_share_link(self, name: str, content_hash: str, proof_code: str, content_hash_name: str, size: str,
                        parent_file_id: str = 'root', force: bool = False) -> bool:
        """
        保存分享文件
        :param content_hash_name:
        :param name:
        :param content_hash:
        :param proof_code:
        :param size:
        :param parent_file_id:
        :param force:
        :return:
        """
        logger.info(f'name: {name}, content_hash:{content_hash}, proof_code:{proof_code}, size:{size}')
        json = {'content_hash': content_hash, 'proof_code': proof_code, 'proof_version': 'v1', 'size': int(size),
                'content_hash_name': content_hash_name or 'sha1'}
        r = self.create_file(name, parent_file_id=parent_file_id, file_type=True, json=json, force=force)
        if r.status_code == 201 and 'rapid_upload' in r.json() and r.json()['rapid_upload']:
            return r.json()['file_id']
        elif 'message' in r.json():
            logger.debug(r.json())
            print(r.json()["message"])
        return False

    def search(self, query: str, raw=False, next_marker: str = None, limit_num: int = 100, limit: bool = False,
               category_list=None):
        """
        搜索文件
        :param query
        :param raw
        :param next_marker
        :param limit_num
        :param limit
        :param category_list
        """
        url = 'https://api.aliyundrive.com/v2/file/search'
        if not raw:
            query = f'name match \"{query}\"'
        if category_list:
            for i in category_list:
                if query:
                    query += ' and '
                query += f'category = \"{i}\"'
        json = {
            'drive_id': self.drive_id,
            'query': query,
            'order_by': 'updated_at DESC',
            'limit': limit_num
        }
        if next_marker:
            json['marker'] = next_marker
        r = self._req.post(url, json=json)
        if 'items' not in r.json():
            return []
        file_list = r.json()['items']
        if 'next_marker' in r.json() and r.json()['next_marker'] and \
                next_marker != r.json()['next_marker'] and not limit:
            file_list.extend(
                self.search(query, raw=True, next_marker=r.json()['next_marker'], limit_num=limit_num))
        return file_list

    def get_play_info(self, file_id, expire_sec=14400, category=None):
        url = 'https://api.aliyundrive.com/v2/databox/get_{}_play_info'
        if category:
            url = url.format(category)
        else:
            return {}
        j = {'drive_id': self.drive_id, 'file_id': file_id, 'expire_sec': expire_sec}
        r = self._req.post(url, json=j)
        if r.status_code != 200:
            return {}
        if 'code' in r.json() and r.json()['code'] == AliyunpanCode.not_found_file:
            return {}
        play_dict = {}
        if 'template_list' in r.json():
            for i in r.json()['template_list']:
                if i['status']:
                    try:
                        play_dict[i['template_id']] = i['url']
                    except KeyError:
                        pass
            return play_dict
        return {}

    def share_link(self, file_id_list: list, expiration=''):
        url = 'https://api.aliyundrive.com/adrive/v2/share_link/create'
        if expiration:
            expiration = datetime.fromtimestamp(float(expiration) + time.timezone) \
                             .strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
        j = {'drive_id': self.drive_id, 'file_id_list': file_id_list, 'expiration': expiration}
        r = self._req.post(url, json=j)
        if 'code' in r.json():
            if r.json()['code'] == AliyunpanCode.Forbidden:
                return ''
            elif r.json()['code'] == AliyunpanCode.InvalidExpiration:
                raise InvalidExpiration
            elif r.json()['code'] == AliyunpanCode.FileShareNotAllowed:
                raise FileShareNotAllowed(r.json()['message'])
            else:
                return ''
        return r.json()['share_url']

    def get_share_by_anonymous(self, share_id: str):
        url = f'https://api.aliyundrive.com/adrive/v3/share_link/get_share_by_anonymous'
        r = self._req.post(url, json={'share_id': share_id})
        return r.json()['file_infos']



import base64
import sys
import time
from pathlib import Path, PurePosixPath

from treelib import Tree
from treelib.exceptions import NodeIDAbsentError

from aliyunpan.api.type import FileInfo, ShareInfo
from aliyunpan.api.utils import get_sha1, get_url_byte, get_proof_code
from aliyunpan.common import GetFileListBar

_all_ = ['PathList', 'parse_share_url', 'AliyunpanPath']


class PathList:
    def __init__(self, disk):
        self._tree = Tree()
        self._disk = disk
        self._tree.create_node(tag='root', identifier='root', data=FileInfo(type=False))
        self.depth = 3

    def update_path_list(self, file_id='root', depth=None, is_fid=True, **kwargs):
        if depth is None:
            depth = self.depth
        kwargs.setdefault('max_depth', depth)
        max_depth = kwargs['max_depth']
        kwargs.setdefault('get_file_list_bar', GetFileListBar(max_depth))
        kwargs.setdefault('ratio', 0)
        get_file_list_bar = kwargs['get_file_list_bar']
        ratio = kwargs['ratio']
        get_file_list_bar.update(refresh_line=False)
        if not is_fid:
            file_id = self.get_path_fid(file_id, update=False)
        file_list = self._disk.get_file_list(file_id)
        if not file_list:
            if depth == max_depth:
                get_file_list_bar.refresh_line()
            return False
        old_file_list = self._tree.children(file_id)
        for i in old_file_list:
            if i.identifier not in [j['file_id'] for j in file_list]:
                self._tree.remove_node(i.identifier)
        for i, info in enumerate(file_list):
            if depth == max_depth:
                ratio = (i + 1) / len(file_list) if file_list else None
            get_file_list_bar.update(depth=max_depth - depth, ratio=ratio, refresh_line=True)
            file_info = self.get_file_info(info)[0]
            if self._tree.get_node(file_info.id):
                self._tree.update_node(file_info.id, data=file_info)
            else:
                self._tree.create_node(tag=file_info.name, identifier=file_info.id, data=file_info, parent=file_id)
            if not file_info.type and depth:
                self.update_path_list(file_id=file_info.id, depth=depth - 1, max_depth=max_depth,
                                      get_file_list_bar=get_file_list_bar, ratio=ratio)
        if depth == max_depth:
            get_file_list_bar.refresh_line()
        return True

    def check_path_diff(self, local_path, disk_path_list):
        p = Path(local_path)
        change_file_list = []
        for path in p.iterdir():
            flag = False
            for i, path_ in enumerate(disk_path_list, 1):
                name, file_info = list(path_.items())[0]
                if p / name not in p.iterdir():
                    change_file_list.append(p / name)
                if Path(path) == p / name:
                    if Path(path).is_dir() and file_info['data'] and path.is_dir() != file_info['data'].type:
                        if 'children' in file_info:
                            children = file_info['children']
                            change_file_list.extend(self.check_path_diff(p / name, children))
                        elif list(path.iterdir()):
                            change_file_list.extend(list(path.iterdir()))
                    if file_info and file_info['data'] and path.is_file() == file_info['data'].type:
                        if path.is_file() and get_sha1(path).lower() != file_info['data'].content_hash.lower():
                            if i == len(disk_path_list):
                                change_file_list.append(path)
                            continue
                        else:
                            flag = True
                if not flag and i == len(disk_path_list):
                    change_file_list.append(path)
        if not len(list(p.iterdir())):
            for path_ in disk_path_list:
                name, file_info = list(path_.items())[0]
                change_file_list.append(p / name)
        if not len(disk_path_list):
            for path_ in p.iterdir():
                change_file_list.append(path_)
        return list(set(change_file_list))

    @staticmethod
    def get_file_info(info):
        file_info_list = []
        if not isinstance(info, list):
            info_list = [info]
        else:
            info_list = info
        for info in info_list:
            if info['type'] == 'file':
                file_info = FileInfo(name=info['name'], id=info['file_id'], pid=info['parent_file_id'], type=True,
                                     ctime=time.strptime(info['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'),
                                     update_time=time.strptime(info['updated_at'], '%Y-%m-%dT%H:%M:%S.%fZ'),
                                     hidden=info['hidden'], category=info['category'],
                                     content_type=info['content_type'],
                                     size=info['size'], content_hash_name=info['content_hash_name'],
                                     content_hash=info['content_hash'],
                                     download_url=info['download_url'] if 'download_url' in info else '',
                                     video_media_metadata=info[
                                         'video_media_metadata'] if 'video_media_metadata' in info else None,
                                     video_preview_metadata=info[
                                         'video_preview_metadata'] if 'video_preview_metadata' in info else None)
            else:
                file_info = FileInfo(name=info['name'], id=info['file_id'], pid=info['parent_file_id'], type=False,
                                     ctime=time.strptime(info['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'),
                                     update_time=time.strptime(info['updated_at'], '%Y-%m-%dT%H:%M:%S.%fZ'),
                                     hidden=info['hidden'])
            file_info_list.append(file_info)
        return file_info_list

    def tree(self, path='root', stdout=sys.stdout):
        file_id = self.get_path_fid(path, update=False)
        self.update_path_list(file_id)
        if not file_id:
            raise FileNotFoundError(path)
        return self._tree.show(file_id, stdout=stdout)

    def get_path_list(self, path, update=True):
        file_id = self.get_path_fid(path, update=update)
        return self.get_fid_list(file_id, update=update)

    def get_fid_list(self, file_id, update=True):
        if not file_id:
            raise FileNotFoundError(Path)
        try:
            self.auto_update_path_list(update, file_id)
        except NodeIDAbsentError:
            return list(map(self.get_file_info, self._disk.get_file_list(file_id)))
        if not self._tree.get_node(file_id):
            return []
        if file_id != 'root' and self._tree.get_node(file_id).data.type:
            return [self._tree.get_node(file_id).data]
        return [i.data for i in self._tree.children(file_id)]

    def get_path_fid(self, path, file_id='root', update=True):
        path = AliyunpanPath(path)
        if str(path) in ('', '/', '\\', '.', 'root'):
            return 'root'
        flag = False
        path_list = list(filter(None, path.split()))
        if path_list[0] == 'root':
            path_list = path_list[1:]
        for i in path_list:
            flag = False
            node_list = self._tree.children(file_id)
            if not node_list:
                self.auto_update_path_list(update, file_id)
                node_list = self._tree.children(file_id)
            for j in node_list:
                if i == j.tag:
                    flag = True
                    file_id = j.identifier
                    break
            if not flag:
                return False
        if flag:
            return file_id
        return False

    def get_path_node(self, path, update=True):
        file_id = self.get_path_fid(path, update=update)
        if file_id:
            return self._tree.get_node(file_id)
        return False

    def get_path_parent_node(self, path, update=True):
        file_id = self.get_path_fid(path, update=update)
        if file_id:
            node = self._tree.parent(file_id)
            if node:
                return node
        return False

    def auto_update_path_list(self, update=True, file_id=None):
        if not update and file_id:
            return self.update_path_list(file_id, depth=0)
        elif update and len(self._tree) == 1:
            return self.update_path_list()


def parse_share_url(url, access_token):
    name, content_hash, url, size, path = url.split('aliyunpan://')[1].split('|')[:5]
    url = base64.b64decode(url).decode()
    split_list = [':', '=']
    content_hash_name = ''
    for i in split_list:
        if i in content_hash:
            content_hash_name, content_hash = content_hash.split(i)
            break
    proof_code = get_proof_code(get_url_byte(url, access_token, int(size)))
    share_info = ShareInfo(name=name, content_hash=content_hash, proof_code=proof_code,
                           content_hash_name=content_hash_name, size=size, path=Path(path.strip()))
    return share_info


class AliyunpanPath(type(Path())):
    def __str__(self):
        path = [i for i in str(PurePosixPath(Path(super().__str__()).as_posix())).split('/') if i != '']
        if not path:
            path = ['root']
        if len(path) != 1 and path[0] == 'root':
            path = path[1:]
        path = '/'.join(path)
        return path

    def split(self):
        return self.__str__().split('/')

    def __eq__(self, other):
        if self.__str__() == other.__str__():
            return True

    def __hash__(self):
        return hash(self.__str__())

    def __sub__(self, other: Path):
        parts = []
        parts1 = self.parts
        parts2 = other.parts
        if len(other.parts) > len(self.parts):
            parts1, parts2 = parts2, parts1
        for i, j in enumerate(parts1):
            if len(parts2) >= i + 1:
                if j != other.parts[i]:
                    parts.append(j)
            else:
                parts.append(j)
        return AliyunpanPath('/'.join(parts))



from pathlib import Path

from ruamel.yaml import YAML

from aliyunpan.common import DATA
from aliyunpan.exceptions import InvalidConfiguration

__all__ = ['Config']

yaml = YAML()


class Config:
    def __init__(self, config_file=None):
        self._config_file = Path(config_file) if config_file else None

    config_file = property(lambda self: self._config_file,
                           lambda self, value: setattr(self, '_config_file', Path(value)))

    def read(self):
        if not self._config_file or not self._config_file.is_file():
            return {}
        with self.config_file.open(encoding='utf-8') as f:
            return yaml.load(f) or {}

    def write(self, conf):
        if not conf:
            try:
                self._config_file.unlink()
            except FileNotFoundError:
                pass
            return
        if not self._config_file or not self._config_file.is_file():
            if not self._config_file.parent.is_dir():
                self._config_file.parent.mkdir(parents=True)
            self._config_file.touch()
        if isinstance(conf, DATA):
            conf = conf.to_dict()
        with self.config_file.open('w', encoding='utf-8') as f:
            yaml.dump(conf, f)
        return True

    def get(self, key):
        if not self._config_file:
            return False
        conf = self.read()
        try:
            if key in conf.keys():
                value = conf.get(key)
                return value
        except AttributeError:
            raise InvalidConfiguration
        return None

    def update(self, key, value):
        if not self._config_file:
            return False
        conf = self.read() or {}
        if isinstance(value, DATA):
            value = dict(value)
        if value is None:
            del conf[key]
        else:
            conf.update({key: value})
        if conf:
            self.write(conf)
        else:
            self._config_file.unlink()
        return conf

    def delete(self, key):
        if not self._config_file:
            return False
        conf = self.read() or {}
        if conf and key in conf:
            conf.pop(key)
            self.write(conf)
        return True



import curses
import functools
import platform
import time
from pathlib import PurePosixPath, Path
from threading import Thread

import npyscreen
import pyperclip

from aliyunpan.api.type import FileInfo
from aliyunpan.api.utils import logger, stop_thread, get_open_port

__all__ = ['AliyunpanTUI']


class Text:
    def __init__(self, text):
        self._text = str(text)

    def __str__(self):
        text = self._text
        if platform.system() == 'Windows':
            text = ''
            for _char in self._text:
                text += _char
                if len(_char.encode()) > 1:
                    text += ' '
        return text

    def __repr__(self):
        text = self._text
        if platform.system() == 'Windows':
            text = ''
            del_list = []
            for i, _char in enumerate(self._text):
                if len(_char.encode()) > 1:
                    if len(self._text) > i + 1 and self._text[i + 1] == ' ':
                        del_list.append(i + 1)
            for i, _char in enumerate(self._text):
                if i not in del_list:
                    text += _char
        return text

    def __eq__(self, other):
        return str(self) == str(other)


class AliyunpanFileForm(npyscreen.FormBaseNewWithMenus):
    def create(self):
        npyscreen.setTheme(npyscreen.Themes.ColorfulTheme)
        self.file_menu = self.add_menu(name='Menu')
        file = self.add(FileGrid)
        self.download_menu = self.file_menu.addNewSubmenu(name='Download Menu', shortcut='^D')
        self.download_menu.addItemsFromList([
            ('Copy the download link', file.copy_download_link, '^C'),
            ('Download file', file.download, '^D'),
            ('Aria2', file.aria2, '^A')
        ])
        self.file_menu.addItemsFromList([
            ('Show file details', file.show_file_info, '^X'),
            ('Cast screen to TV', file.dlna, '^A'),
            ('Search', file.search, '^F'),
            ('Remove', file.remove, '^R'),
            ('Mkdir', file.mkdir, 'm')
        ])


class Select(npyscreen.TitleSelectOne):
    def set_up_handlers(self):
        super(Select, self).set_up_handlers()
        self.handlers.update({
            '^C': self._exit
        })

    def _exit(self, _):
        raise KeyboardInterrupt


class DeviceSelect(Select):
    pass


class QualitySelect(Select):
    pass


class Time(npyscreen.TitleText):
    def set_up_handlers(self):
        super(Time, self).set_up_handlers()
        self.handlers.update({
            '^C': self._exit
        })

    def _exit(self, _):
        raise KeyboardInterrupt


class Dlna(npyscreen.FormWithMenus):
    def create(self):
        self.device_menu = self.add_menu(name='Device Menu')
        self.device_menu.addItemsFromList([
            ('Play', self.play, '^A'),
            ('Proxy Play', self.proxy_play, 'y'),
            ('Local redirection Play', self.redirect_play, 'r'),
            ('Pause', self.pause, 'p'),
            ('Continue', self.continue_, 'c'),
            ('Mute', self.mute, '^Q'),
            ('Unmute', self.unmute, '^E'),
            ('Stop', self.stop, '^S'),
            ('Back', self.back, '^X'),
            ('Refresh Device', self.refresh_device, '^R')
        ])
        self.devices = []
        self.quality_dict = {}
        self.quality = 'default'
        self.device_select = self.add(DeviceSelect, scroll_exit=False, name='Devices', max_height=5,
                                      value_changed_callback=self.device_changed_callback)
        self.quality_select = self.add(QualitySelect, scroll_exit=True, name='Quality', max_height=5,
                                       value_changed_callback=self.quality_changed_callback)
        self.add(npyscreen.TitleSlider, color='STANDOUT', name='Volume', out_of=10,
                 value_changed_callback=self.volume_changed_callback)
        self.position = '00:00:00'
        self.second = self.add(Time, name='sec', value='00')
        self.minute = self.add(Time, name='min', value='00')
        self.hour = self.add(Time, name='hour', value='00')
        self._volume = 0
        self._proxy_port = 8000
        self._change_lock_time = 0
        self.add(npyscreen.ButtonPress, relx=0, name='Set time', when_pressed_function=self.set_time)
        self._proxy_thread = None
        self._redirect_thread = None
        self.proxy = None
        self.redirect = None

    def back(self):
        self.parentApp.switchFormPrevious()

    def beforeEditing(self):
        self.file_info = self.parentApp.file_info
        self.name = str(Text(self.file_info.name))
        try:
            from dlnap.dlnap import dlnap
        except ImportError as e:
            self.name = e.__str__()
            logger.error('Cannot find submodule dlnap.')
        else:
            self.dlnap = dlnap
            self._refresh()

    def afterEditing(self):
        self.parentApp.setNextFormPrevious()

    def device_changed_callback(self, widget):
        if time.time() - self._change_lock_time > 1:
            self._change_lock_time = 0
        if widget.value and len(self.devices) and not self._change_lock_time:
            self._change_lock_time = time.time()
            device = self.devices[widget.value[0]]
            if self.quality and self.quality_dict:
                url = self.quality_dict[self.quality]
            else:
                url = self.parentApp._cli._disk.get_download_url(self.file_info.id)
            logger.info(f'Device {device} is playing {self.file_info.name}')
            if url:
                if self.proxy:
                    ip = self.dlnap._get_serve_ip(device.ip)
                    if self._proxy_thread:
                        stop_thread(self._proxy_thread)
                    self._proxy_thread = Thread(target=self.dlnap.runProxy, daemon=True,
                                                kwargs={'ip': ip, 'port': get_open_port()})
                    self._proxy_thread.start()
                    url = 'http://{}:{}/{}'.format(ip, self._proxy_port, url)
                elif self.redirect:
                    ip = self.dlnap._get_serve_ip(device.ip)
                    if self._redirect_thread:
                        stop_thread(self._redirect_thread)
                    port = get_open_port()
                    self._redirect_thread = Thread(target=Dlna._redirect, args=(url, port), daemon=True).start()
                    url = 'http://{}:{}/'.format(ip, port)
                logger.debug(url)
                device.set_current_media(url)
                device.play()

    def quality_changed_callback(self, widget):
        if widget.value:
            try:
                self.quality = list(self.quality_dict.keys())[widget.value[0]]
            except IndexError:
                self.quality = 'default'
            self.play()

    def play(self, proxy=False, redirect=False):
        if proxy:
            self.proxy = True
            self.redirect = False
        elif redirect:
            self.redirect = True
            self.proxy = False
        if len(self.devices):
            if self.device_select.value:
                self.device_changed_callback(self.device_select)
            else:
                self.device_select.value = [self.device_select.entry_widget.cursor_line]
                self.device_changed_callback(self.device_select)

    def proxy_play(self):
        self.play(proxy=True)

    @staticmethod
    def _redirect(url, port):
        from flask import Flask, redirect

        app = Flask(__name__)

        @app.route('/')
        def index():
            return redirect(url, code=301)

        app.run(host='0.0.0.0', port=port)

    def redirect_play(self):
        self.play(redirect=True)

    def pause(self):
        if len(self.devices) and self.device_select.entry_widget.value:
            device = self.devices[self.device_select.entry_widget.value[0]]
            device.pause()

    def continue_(self):
        if len(self.devices) and self.device_select.entry_widget.value:
            device = self.devices[self.device_select.entry_widget.value[0]]
            device.play()

    def stop(self):
        if len(self.devices) and self.device_select.entry_widget.value:
            device = self.devices[self.device_select.entry_widget.value[0]]
            self.device_select.entry_widget.value = []
            device.stop()

    def mute(self):
        if len(self.devices) and self.device_select.entry_widget.value:
            device = self.devices[self.device_select.entry_widget.value[0]]
            device.mute()

    def unmute(self):
        if len(self.devices) and self.device_select.entry_widget.value:
            device = self.devices[self.device_select.entry_widget.value[0]]
            device.unmute()

    def discover(self, display=True):
        self.name = 'Refresh Device...'
        allDevices = self.dlnap.discover(st=self.dlnap.URN_AVTransport)
        self.name = str(Text(self.file_info.name))
        for device in allDevices:
            if device not in self.devices:
                self.devices.append(device)
        self.device_select.values = [Text(i) for i in self.devices]
        logger.debug(self.devices)
        if display:
            self.display()

    def refresh_device(self, display=True):
        t = Thread(target=Dlna.discover, args=(self, display), daemon=False)
        t.start()
        return t

    def refresh_quality(self, display=True):
        t = Thread(target=Dlna.get_quality_info, args=(self, display), daemon=False)
        t.start()
        return t

    def _refresh(self):
        def t():
            self.refresh_device(display=False).join()
            self.refresh_quality(display=False).join()
            if self.editing:
                self.display()

        Thread(target=t).start()

    def get_quality_info(self, display=True):
        self.quality_dict = {'default': self.parentApp._cli._disk.get_download_url(self.file_info.id)}
        self.quality_dict.update(
            self.parentApp._cli._disk.get_play_info(file_id=self.file_info.id, category=self.file_info.category))
        self.quality_select.values = list(self.quality_dict.keys())
        logger.debug(self.quality_dict)
        if display:
            self.display()

    def volume_changed_callback(self, widget):
        if len(self.devices) and self.device_select.entry_widget.value and self._volume != widget.value:
            self._volume = widget.value
            device = self.devices[self.device_select.entry_widget.value[0]]
            device.volume(int(widget.value) * 10)

    def set_position(self, second=None, minute=None, hour=None):
        position = '{:02d}:{:02d}:{:02d}'
        if len(self.devices) and self.device_select.entry_widget.value:
            position = position.format(hour, minute, second)
            npyscreen.notify_confirm(position)
            device = self.devices[self.device_select.entry_widget.value[0]]
            device.seek(position)

    def set_time(self):
        try:
            self.set_position(second=int(self.second.value), minute=int(self.minute.value), hour=int(self.hour.value))
        except ValueError:
            pass


class FileGrid(npyscreen.SimpleGrid):
    def __init__(self, *args, **kwargs):
        super(FileGrid, self).__init__(*args, **kwargs)
        self._file_list = []
        self._parent_file_info = None
        self.parent.name = 'root'
        self.update_file_list(file_id='root')

    file_name = property(lambda self: self.values[self.edit_cell[0]][self.edit_cell[1]])
    _file_info = property(lambda self: [file_info for file_info in self._file_list if file_info.name == self.file_name])
    file_info = property(lambda self: self._file_info[0] if self._file_info else None)
    parent_file_info = property(lambda self: self._parent_file_info)

    def set_up_handlers(self):
        super(FileGrid, self).set_up_handlers()
        self.handlers.update({
            curses.ascii.NL: self.keyboard_handlers,
            curses.ascii.CR: self.keyboard_handlers,
            curses.ascii.SP: self.keyboard_handlers,
            curses.ascii.ESC: self.back,
            '^C': self._exit
        })

    def back(self, _):
        if self._parent_file_info:
            self.update_file_list(file_id=self._parent_file_info.pid)

    def _exit(self, _):
        raise KeyboardInterrupt

    def download(self, aria2=False):
        if self.file_info:
            path = Path(Text(self.parent.name).__repr__()) / Path(self.file_info.name)
            if self.file_info.type:
                if aria2:
                    Thread(target=functools.partial(self.parent.parentApp._cli.download, str(path), aria2=True)).start()
                else:
                    Thread(target=functools.partial(self.parent.parentApp._cli.download_file,
                                                    path=Path(self.file_info.name),
                                                    url=self.parent.parentApp._cli._disk.get_download_url(
                                                        self.file_info.id))).start()
            else:
                if self.parent.name == 'root':
                    path = Path(self.file_info.name)
                if aria2:
                    Thread(target=functools.partial(self.parent.parentApp._cli.download, str(path), aria2=True)).start()
                else:
                    Thread(target=functools.partial(self.parent.parentApp._cli.download, path=str(path))).start()

    def aria2(self):
        self.download(aria2=True)

    def copy_download_link(self):
        if self.file_info and self.file_info.name == self.file_name and self.file_info.type:
            url = self.parent.parentApp._cli._disk.get_download_url(self.file_info.id)
            if url:
                pyperclip.copy(url)
                npyscreen.notify_confirm(f'Already copied to clipboard!\n{Text(self.file_info.name)}\n{url}')

    def dlna(self):
        if self.file_info and self.file_info.name == self.file_name and self.file_info.type:
            self.parent.parentApp.file_info = self.file_info
            self.parent.parentApp.switchForm('DLNA')

    def search(self):
        self.parent.parentApp.file_grid = self
        self.parent.parentApp.switchForm('Search')

    def remove(self):
        if self.file_info and Text(self.file_info.name) == Text(self.file_name):
            self.parent.parentApp._cli.rm(path=None, file_id=self.file_info.id)
            file_list = []
            for i in self.values:
                file_list.extend(i)
            file_list.remove(self.file_info.name)
            self.set_grid_values_from_flat_list(file_list)

    def mkdir(self):
        self.parent.parentApp.file_grid = self
        self.parent.parentApp.switchForm('Mkdir')

    def show_file_info(self, file_info=None):
        if not file_info:
            file_info = self.file_info
        info = f'name:{Text(file_info.name)}\n' \
               f'file_id:{file_info.id}\n' \
               f'type:{"file" if file_info.type else "folder"}\n' \
               f'create_time:{time.strftime("%d %b %H:%M", file_info.ctime)}\n' \
               f'update_time:{time.strftime("%d %b %H:%M", file_info.update_time)}\n' \
               f'category:{file_info.category}\n' \
               f'content_type:{file_info.content_type}\n' \
               f'size:{file_info.size}\n' \
               f'content_hash_name:{file_info.content_hash_name}\n' \
               f'content_hash:{file_info.content_hash}'
        npyscreen.notify_confirm(info, 'File info')

    def update_file_list(self, name=None, file_id=None):
        file_list = []
        self.searched = False
        if name == '..' and self._parent_file_info and self._parent_file_info.pid:
            # 返回父目录且有父目录
            return self.update_file_list(file_id=self._parent_file_info.pid)
        elif name:
            # 进入目录name
            file_info = self.file_info
            if file_info:
                if file_info.type:
                    self.show_file_info(file_info)
                    return
                self._file_list = self.parent.parentApp._cli._path_list.get_fid_list(file_info.id, update=False)
                # 保存当前目录信息
                self._parent_file_info = file_info
        elif file_id:
            # 进入目录file_id
            self._file_list = self.parent.parentApp._cli._path_list.get_fid_list(file_id, update=False)
            # 保存当前目录信息
            self._parent_file_info = None
            if file_id != 'root':
                self._parent_file_info = self.parent.parentApp._cli._path_list._tree.get_node(file_id).data
        else:
            self._file_list = self.parent.parentApp._cli._path_list.get_fid_list(self._parent_file_info.id,
                                                                                 update=False)
        for file_info in self._file_list:
            file_list.append(file_info.name)
        if self._parent_file_info and self._parent_file_info.pid:
            # 有父目录
            file_list.insert(0, '..')
        if file_list:
            self.set_grid_values_from_flat_list(file_list)
        self.update_path()

    def update_path(self):
        path = 'root'
        if self._parent_file_info:
            path = PurePosixPath(self._parent_file_info.name)
        pid = self._parent_file_info.pid if self._parent_file_info else None
        while True:
            if pid:
                file_info = self.parent.parentApp._cli._path_list._tree.get_node(pid).data
                if file_info:
                    path = file_info.name / path
                    pid = file_info.pid
                else:
                    pid = None
            else:
                break
        if path:
            self.parent.name = str(Text(path))
            self._display()

    def _display(self):
        if platform.system() == 'Windows':
            self.parent.display()
        else:
            self.parent.DISPLAY()

    def keyboard_handlers(self, _):
        self.update_file_list(self.file_name)

    def custom_print_cell(self, actual_cell, cell_display_value):
        pass

    def display_value(self, vl):
        return Text(vl)


class Search(npyscreen.ActionPopup):
    def create(self):
        self.query = self.add(npyscreen.TitleText, name='query')
        self.limit_num = self.add(npyscreen.TitleText, name='limit num', value='100')
        self.limit = self.add(npyscreen.Checkbox, name='limit', value=False)
        self.raw = self.add(npyscreen.Checkbox, name='raw')
        self.category_list = []
        self.category = self.add(npyscreen.TitleMultiSelect, name='category', scroll_exit=True,
                                 value_changed_callback=lambda widget: setattr(self, 'category_list',
                                                                               [self.category.values[i] for i in
                                                                                widget.value]))
        self.category.values = ['video', 'audio', 'image', 'doc', 'zip', 'others']

    def afterEditing(self):
        self.parentApp.setNextFormPrevious()

    def on_ok(self):
        if not self.query.value:
            self.raw.value = True
        file_info_list = self.parentApp._cli._path_list.get_file_info(
            self.parentApp._cli._disk.search(self.query.value, raw=self.raw.value, limit_num=int(self.limit_num.value),
                                             limit=self.limit.value, category_list=self.category_list))
        name = f'{time.time()} - {self.query.value}'
        self.parentApp._cli._path_list._tree.create_node(tag=name, identifier=name, parent='root',
                                                         data=FileInfo(name, name, 'root', False, time.time(),
                                                                       time.time()))
        for i in file_info_list:
            self.parentApp._cli._path_list._tree.create_node(tag=i.name, identifier=i.id, parent=name, data=i)
        if not self.parentApp.file_grid.searched:
            self.parentApp.file_grid.searched = True
            self.parentApp.file_grid._parent_file_info = self.parentApp.file_grid.file_info
        self.parentApp.file_grid._file_list = file_info_list
        file_list = ['..']
        for file_info in file_info_list:
            file_list.append(file_info.name)
        self.parentApp.file_grid.set_grid_values_from_flat_list(file_list)
        self.parentApp.file_grid.parent.name = name
        self.parentApp.file_grid._display()

    def on_cancel(self):
        pass


class Mkdir(npyscreen.ActionPopup):
    def create(self):
        self.name = self.add(npyscreen.TitleText, name='name')

    def afterEditing(self):
        self.parentApp.setNextFormPrevious()

    def on_ok(self):
        if not self.parentApp.file_grid.searched:
            if self.parentApp.file_grid.parent_file_info:
                parent_file_id = self.parentApp.file_grid.parent_file_info.id
            else:
                parent_file_id = 'root'
            file_id = self.parentApp._cli.mkdir(path=None, name=self.name.value, parent_file_id=parent_file_id)[0][0]
            file_list = []
            for i in self.parentApp.file_grid.values:
                file_list.extend(i)
            file_list.append(str(Text(self.name.value)))
            self.parentApp.file_grid._file_list.append(self.parentApp._cli._path_list._tree.get_node(file_id).data)
            self.parentApp.file_grid.set_grid_values_from_flat_list(file_list)

    def on_cancel(self):
        pass


class AliyunpanTUI(npyscreen.NPSAppManaged):
    def __init__(self, cli):
        self._cli = cli
        self.file_info = None
        super(AliyunpanTUI, self).__init__()

    def onStart(self):
        self.addForm('MAIN', AliyunpanFileForm)
        self.addForm('DLNA', Dlna)
        self.addForm('Search', Search)
        self.addForm('Mkdir', Mkdir)



from collections import namedtuple

__all__ = ['FileInfo', 'UserInfo', 'ShareInfo', 'AlibumInfo']

_file_info = (
    'name', 'id', 'pid', 'type', 'ctime', 'update_time', 'hidden', 'category', 'content_type', 'size',
    'content_hash_name',
    'content_hash', 'download_url',
    'video_media_metadata', 'video_preview_metadata')
FileInfo = namedtuple('FileInfo', _file_info)
FileInfo.__new__.__defaults__ = ('',) * len(_file_info)
UserInfo = namedtuple('UserInfo', ['id', 'nick_name', 'ctime', 'phone', 'drive_id'])
UserInfo.__new__.__defaults__ = ('',) * 5
ShareInfo = namedtuple('ShareInfo', ['name', 'content_hash', 'proof_code', 'content_hash_name', 'size', 'path'])
ShareInfo.__new__.__defaults__ = ('',) * 6
AlibumInfo = namedtuple('AlibumInfo', ['drive_name', 'drive_id'])
AlibumInfo.__new__.__defaults__ = ('',) * 2



#!/usr/bin/python

# @file dlnap.py
# @author cherezov.pavel@gmail.com
# @brief Python over the network media player to playback on DLNA UPnP devices.

# Change log:
#   0.1  initial version.
#   0.2  device renamed to DlnapDevice; DLNAPlayer is disappeared.
#   0.3  debug output is added. Extract location url fixed.
#   0.4  compatible discover mode added.
#   0.5  xml parser introduced for device descriptions
#   0.6  xpath introduced to navigate over xml dictionary
#   0.7  device ip argument introduced
#   0.8  debug output is replaced with standard logging
#   0.9  pause/stop added. Video playback tested on Samsung TV
#   0.10 proxy (draft) is introduced.
#   0.11 sync proxy for py2 and py3 implemented, --proxy-port added
#   0.12 local files can be played as well now via proxy
#   0.13 ssdp protocol version argument added
#   0.14 fixed bug with receiving responses from device
#   0.15 Lot's of fixes and features added thanks @ttopholm and @NicoPy
#
#   1.0  moved from idea version

__version__ = "0.15"

import re
import sys
import time
import signal
import socket
import select
import logging
import traceback
import mimetypes
from contextlib import contextmanager


import os
py3 = sys.version_info[0] == 3
if py3:
    from urllib.request import urlopen
    from http.server import HTTPServer
    from http.server import BaseHTTPRequestHandler
else:
    from urllib2 import urlopen
    from BaseHTTPServer import BaseHTTPRequestHandler
    from BaseHTTPServer import HTTPServer

import shutil
import threading

SSDP_GROUP = ("239.255.255.250", 1900)
URN_AVTransport = "urn:schemas-upnp-org:service:AVTransport:1"
URN_AVTransport_Fmt = "urn:schemas-upnp-org:service:AVTransport:{}"

URN_RenderingControl = "urn:schemas-upnp-org:service:RenderingControl:1"
URN_RenderingControl_Fmt = "urn:schemas-upnp-org:service:RenderingControl:{}"

SSDP_ALL = "ssdp:all"

# =================================================================================================
# XML to DICT
#
def _get_tag_value(x, i = 0):
   """ Get the nearest to 'i' position xml tag name.

   x -- xml string
   i -- position to start searching tag from
   return -- (tag, value) pair.
      e.g
         <d>
            <e>value4</e>
         </d>
      result is ('d', '<e>value4</e>')
   """
   x = x.strip()
   value = ''
   tag = ''

   # skip <? > tag
   if x[i:].startswith('<?'):
      i += 2
      while i < len(x) and x[i] != '<':
         i += 1

   # check for empty tag like '</tag>'
   if x[i:].startswith('</'):
      i += 2
      in_attr = False
      while i < len(x) and x[i] != '>':
         if x[i] == ' ':
            in_attr = True
         if not in_attr:
            tag += x[i]
         i += 1
      return (tag.strip(), '', x[i+1:])

   # not an xml, treat like a value
   if not x[i:].startswith('<'):
      return ('', x[i:], '')

   i += 1 # <

   # read first open tag
   in_attr = False
   while i < len(x) and x[i] != '>':
      # get rid of attributes
      if x[i] == ' ':
         in_attr = True
      if not in_attr:
         tag += x[i]
      i += 1

   i += 1 # >
   
   # replace self-closing <tag/> by <tag>None</tag> 
   empty_elmt = '<' + tag + ' />'
   closed_elmt = '<' + tag + '>None</'+tag+'>'
   if x.startswith(empty_elmt):
      x = x.replace(empty_elmt, closed_elmt)

   while i < len(x):
      value += x[i]
      if x[i] == '>' and value.endswith('</' + tag + '>'):
         # Note: will not work with xml like <a> <a></a> </a>
         close_tag_len = len(tag) + 2 # />
         value = value[:-close_tag_len]
         break
      i += 1
   return (tag.strip(), value[:-1], x[i+1:])

def _xml2dict(s, ignoreUntilXML = False):

   """ Convert xml to dictionary.

   <?xml version="1.0"?>
   <a any_tag="tag value">
      <b> <bb>value1</bb> </b>
      <b> <bb>value2</bb> </b>
      </c>
      <d>
         <e>value4</e>
      </d>
      <g>value</g>
   </a>

   =>

   { 'a':
     {
         'b': [ {'bb':value1}, {'bb':value2} ],
         'c': [],
         'd':
         {
           'e': [value4]
         },
         'g': [value]
     }
   }
   """
   if ignoreUntilXML:
      s = ''.join(re.findall(".*?(<.*)", s, re.M))

   d = {}
   while s:
      tag, value, s = _get_tag_value(s)
      value = value.strip()
      isXml, dummy, dummy2 = _get_tag_value(value)
      if tag not in d:
         d[tag] = []
      if not isXml:
         if not value:
            continue
         d[tag].append(value.strip())
      else:
         if tag not in d:
            d[tag] = []
         d[tag].append(_xml2dict(value))
   return d

s = """
   hello
   this is a bad
   strings

   <?xml version="1.0"?>
   <a any_tag="tag value">
      <b><bb>value1</bb></b>
      <b><bb>value2</bb> <v>value3</v></b>
      </c>
      <d>
         <e>value4</e>
      </d>
      <g>value</g>
   </a>
"""

def _xpath(d, path):
   """ Return value from xml dictionary at path.

   d -- xml dictionary
   path -- string path like root/device/serviceList/service@serviceType=URN_AVTransport/controlURL
   return -- value at path or None if path not found
   """

   for p in path.split('/'):
      tag_attr = p.split('@')
      tag = tag_attr[0]
      if tag not in d:
         return None

      attr = tag_attr[1] if len(tag_attr) > 1 else ''
      if attr:
         a, aval = attr.split('=')
         for s in d[tag]:
            if s[a] == [aval]:
               d = s
               break
      else:
         d = d[tag][0]
   return d
#
# XML to DICT
# =================================================================================================
# PROXY
#
running = False
class DownloadProxy(BaseHTTPRequestHandler):

   def log_message(self, format, *args):
      pass

   def log_request(self, code='-', size='-'):
      pass

   def response_success(self):
      url = self.path[1:] # replace '/'

      if os.path.exists(url):
         f = open(url)
         content_type = mimetypes.guess_type(url)[0]
      else:
         f = urlopen(url=url)

         if py3:
            content_type = f.getheader("Content-Type")
         else:
            content_type = f.info().getheaders("Content-Type")[0]

      self.send_response(200, "ok")
      self.send_header('Access-Control-Allow-Origin', '*')
      self.send_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
      self.send_header("Access-Control-Allow-Headers", "X-Requested-With")
      self.send_header("Access-Control-Allow-Headers", "Content-Type")
      self.send_header("Content-Type", content_type)
      self.end_headers()

   def do_OPTIONS(self):
      self.response_success()

   def do_HEAD(self):
      self.response_success()

   def do_GET(self):
      global running
      url = self.path[1:] # replace '/'

      content_type = ''
      if os.path.exists(url):
         f = open(url)
         content_type = mimetypes.guess_type(url)[0]
         size = os.path.getsize(url)
      elif not url or not url.startswith('http'):
         self.response_success()
         return
      else:
         f = urlopen(url=url)

      try:
         if not content_type:
            if py3:
               content_type = f.getheader("Content-Type")
               size = f.getheader("Content-Length")
            else:
               content_type = f.info().getheaders("Content-Type")[0]
               size = f.info().getheaders("Content-Length")[0]

         self.send_response(200)
         self.send_header('Access-Control-Allow-Origin', '*')
         self.send_header("Content-Type", content_type)
         self.send_header("Content-Disposition", 'attachment; filename="{}"'.format(os.path.basename(url)))
         self.send_header("Content-Length", str(size))
         self.end_headers()
         shutil.copyfileobj(f, self.wfile)
      finally:
         running = False
         f.close()

def runProxy(ip = '', port = 8000):
   global running
   running = True
   DownloadProxy.protocol_version = "HTTP/1.0"
   httpd = HTTPServer((ip, port), DownloadProxy)
   while running:
      httpd.handle_request()

#
# PROXY
# =================================================================================================

def _get_port(location):
   """ Extract port number from url.

   location -- string like http://anyurl:port/whatever/path
   return -- port number
   """
   port = re.findall('http://.*?:(\d+).*', location)
   return int(port[0]) if port else 80


def _get_control_url(xml, urn):
   """ Extract AVTransport contol url from device description xml

   xml -- device description xml
   return -- control url or empty string if wasn't found
   """
   return _xpath(xml, 'root/device/serviceList/service@serviceType={}/controlURL'.format(urn))

@contextmanager
def _send_udp(to, packet):
   """ Send UDP message to group

   to -- (host, port) group to send the packet to
   packet -- message to send
   """
   sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
   sock.sendto(packet.encode(), to)
   yield sock
   sock.close()

def _unescape_xml(xml):
   """ Replace escaped xml symbols with real ones.
   """
   return xml.replace('&lt;', '<').replace('&gt;', '>').replace('&quot;', '"')

def _send_tcp(to, payload):
   """ Send TCP message to group

   to -- (host, port) group to send to payload to
   payload -- message to send
   """
   try:
      sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      sock.settimeout(5)
      sock.connect(to)
      sock.sendall(payload.encode('utf-8'))

      data = sock.recv(2048)
      if py3:
         data = data.decode('utf-8')
      data = _xml2dict(_unescape_xml(data), True)

      errorDescription = _xpath(data, 's:Envelope/s:Body/s:Fault/detail/UPnPError/errorDescription')
      if errorDescription is not None:
         logging.error(errorDescription)
   except Exception as e:
      data = ''
   finally:
      sock.close()
   return data


def _get_location_url(raw):
    """ Extract device description url from discovery response

    raw -- raw discovery response
    return -- location url string
    """
    t = re.findall('\n(?i)location:\s*(.*)\r\s*', raw, re.M)
    if len(t) > 0:
        return t[0]
    return ''

def _get_friendly_name(xml):
   """ Extract device name from description xml

   xml -- device description xml
   return -- device name
   """
   name = _xpath(xml, 'root/device/friendlyName')
   return name if name is not None else 'Unknown'

def _get_serve_ip(target_ip, target_port=80):
    """ Find ip address of network interface used to communicate with target
    
    target-ip -- ip address of target
    return -- ip address of interface connected to target
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect((target_ip, target_port))
    my_ip = s.getsockname()[0]
    s.close()
    return my_ip

class DlnapDevice:
   """ Represents DLNA/UPnP device.
   """

   def __init__(self, raw, ip):
      self.__logger = logging.getLogger(self.__class__.__name__)
      self.__logger.info('=> New DlnapDevice (ip = {}) initialization..'.format(ip))

      self.ip = ip
      self.ssdp_version = 1

      self.port = None
      self.name = 'Unknown'
      self.control_url = None
      self.rendering_control_url = None
      self.has_av_transport = False

      try:
         self.__raw = raw.decode()
         self.location = _get_location_url(self.__raw)
         self.__logger.info('location: {}'.format(self.location))

         self.port = _get_port(self.location)
         self.__logger.info('port: {}'.format(self.port))

         raw_desc_xml = urlopen(self.location).read().decode()

         self.__desc_xml = _xml2dict(raw_desc_xml)
         self.__logger.debug('description xml: {}'.format(self.__desc_xml))

         self.name = _get_friendly_name(self.__desc_xml)
         self.__logger.info('friendlyName: {}'.format(self.name))

         self.control_url = _get_control_url(self.__desc_xml, URN_AVTransport)
         self.__logger.info('control_url: {}'.format(self.control_url))

         self.rendering_control_url = _get_control_url(self.__desc_xml, URN_RenderingControl)
         self.__logger.info('rendering_control_url: {}'.format(self.rendering_control_url))

         self.has_av_transport = self.control_url is not None
         self.__logger.info('=> Initialization completed'.format(ip))
      except Exception as e:
         self.__logger.warning('DlnapDevice (ip = {}) init exception:\n{}'.format(ip, traceback.format_exc()))

   def __repr__(self):
      return '{} @ {}'.format(self.name, self.ip)

   def __eq__(self, d):
      return self.name == d.name and self.ip == d.ip

   def _payload_from_template(self, action, data, urn):
      """ Assembly payload from template.
      """
      fields = ''
      for tag, value in data.items():
        fields += '<{tag}>{value}</{tag}>'.format(tag=tag, value=value)

      payload = """<?xml version="1.0" encoding="utf-8"?>
         <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
            <s:Body>
               <u:{action} xmlns:u="{urn}">
                  {fields}
               </u:{action}>
            </s:Body>
         </s:Envelope>""".format(action=action, urn=urn, fields=fields)
      return payload

   def _create_packet(self, action, data):
      """ Create packet to send to device control url.

      action -- control action
      data -- dictionary with XML fields value
      """
      if action in ["SetVolume", "SetMute", "GetVolume"]:
          url = self.rendering_control_url
          urn = URN_RenderingControl_Fmt.format(self.ssdp_version)
      else:
          url = self.control_url
          urn = URN_AVTransport_Fmt.format(self.ssdp_version)
      payload = self._payload_from_template(action=action, data=data, urn=urn)

      packet = "\r\n".join([
         'POST {} HTTP/1.1'.format(url),
         'User-Agent: {}/{}'.format(__file__, __version__),
         'Accept: */*',
         'Content-Type: text/xml; charset="utf-8"',
         'HOST: {}:{}'.format(self.ip, self.port),
         'Content-Length: {}'.format(len(payload)),
         'SOAPACTION: "{}#{}"'.format(urn, action),
         'Connection: close',
         '',
         payload,
         ])

      self.__logger.debug(packet)
      return packet

   def set_current_media(self, url, instance_id = 0):
      """ Set media to playback.

      url -- media url
      instance_id -- device instance id
      """
      packet = self._create_packet('SetAVTransportURI', {'InstanceID':instance_id, 'CurrentURI':url, 'CurrentURIMetaData':'' })
      _send_tcp((self.ip, self.port), packet)

   def play(self, instance_id = 0):
      """ Play media that was already set as current.

      instance_id -- device instance id
      """
      packet = self._create_packet('Play', {'InstanceID': instance_id, 'Speed': 1})
      _send_tcp((self.ip, self.port), packet)

   def pause(self, instance_id = 0):
      """ Pause media that is currently playing back.

      instance_id -- device instance id
      """
      packet = self._create_packet('Pause', {'InstanceID': instance_id, 'Speed':1})
      _send_tcp((self.ip, self.port), packet)

   def stop(self, instance_id = 0):
      """ Stop media that is currently playing back.

      instance_id -- device instance id
      """
      packet = self._create_packet('Stop', {'InstanceID': instance_id, 'Speed': 1})
      _send_tcp((self.ip, self.port), packet)


   def seek(self, position, instance_id = 0):
      """
      Seek position
      """
      packet = self._create_packet('Seek', {'InstanceID':instance_id, 'Unit':'REL_TIME', 'Target': position })
      _send_tcp((self.ip, self.port), packet)


   def volume(self, volume=10, instance_id = 0):
      """ Stop media that is currently playing back.

      instance_id -- device instance id
      """
      packet = self._create_packet('SetVolume', {'InstanceID': instance_id, 'DesiredVolume': volume, 'Channel': 'Master'})

      _send_tcp((self.ip, self.port), packet)
      
      
   def get_volume(self, instance_id = 0):
      """
      get volume
      """
      packet = self._create_packet('GetVolume', {'InstanceID':instance_id, 'Channel': 'Master'})
      _send_tcp((self.ip, self.port), packet)


   def mute(self, instance_id = 0):
      """ Stop media that is currently playing back.

      instance_id -- device instance id
      """
      packet = self._create_packet('SetMute', {'InstanceID': instance_id, 'DesiredMute': '1', 'Channel': 'Master'})
      _send_tcp((self.ip, self.port), packet)

   def unmute(self, instance_id = 0):
      """ Stop media that is currently playing back.

      instance_id -- device instance id
      """
      packet = self._create_packet('SetMute', {'InstanceID': instance_id, 'DesiredMute': '0', 'Channel': 'Master'})
      _send_tcp((self.ip, self.port), packet)

   def info(self, instance_id=0):
      """ Transport info.

      instance_id -- device instance id
      """
      packet = self._create_packet('GetTransportInfo', {'InstanceID': instance_id})
      return _send_tcp((self.ip, self.port), packet)

   def media_info(self, instance_id=0):
      """ Media info.

      instance_id -- device instance id
      """
      packet = self._create_packet('GetMediaInfo', {'InstanceID': instance_id})
      return _send_tcp((self.ip, self.port), packet)


   def position_info(self, instance_id=0):
      """ Position info.
      instance_id -- device instance id
      """
      packet = self._create_packet('GetPositionInfo', {'InstanceID': instance_id})
      return _send_tcp((self.ip, self.port), packet)


   def set_next(self, url):
      pass

   def next(self):
      pass


def discover(name = '', ip = '', timeout = 1, st = SSDP_ALL, mx = 3, ssdp_version = 1):
   """ Discover UPnP devices in the local network.

   name -- name or part of the name to filter devices
   timeout -- timeout to perform discover
   st -- st field of discovery packet
   mx -- mx field of discovery packet
   return -- list of DlnapDevice
   """
   st = st.format(ssdp_version)
   payload = "\r\n".join([
              'M-SEARCH * HTTP/1.1',
              'User-Agent: {}/{}'.format(__file__, __version__),
              'HOST: {}:{}'.format(*SSDP_GROUP),
              'Accept: */*',
              'MAN: "ssdp:discover"',
              'ST: {}'.format(st),
              'MX: {}'.format(mx),
              '',
              ''])
   devices = []
   with _send_udp(SSDP_GROUP, payload) as sock:
      start = time.time()
      while True:
         if time.time() - start > timeout:
            # timed out
            break
         r, w, x = select.select([sock], [], [sock], 1)
         if sock in r:
             data, addr = sock.recvfrom(1024)
             if ip and addr[0] != ip:
                continue

             d = DlnapDevice(data, addr[0])
             d.ssdp_version = ssdp_version
             if d not in devices:
                if not name or name is None or name.lower() in d.name.lower():
                   if not ip:
                      devices.append(d)
                   elif d.has_av_transport:
                      # no need in further searching by ip
                      devices.append(d)
                      break

         elif sock in x:
             raise Exception('Getting response failed')
         else:
             # Nothing to read
             pass
   return devices

#
# Signal of Ctrl+C
# =================================================================================================
def signal_handler(signal, frame):
   print(' Got Ctrl + C, exit now!')
   sys.exit(1)

signal.signal(signal.SIGINT, signal_handler)

if __name__ == '__main__':
   import getopt

   def usage():
      print('{} [--ip <device ip>] [-d[evice] <name>] [--all] [-t[imeout] <seconds>] [--play <url>] [--pause] [--stop] [--proxy]'.format(__file__))
      print(' --ip <device ip> - ip address for faster access to the known device')
      print(' --device <device name or part of the name> - discover devices with this name as substring')
      print(' --all - flag to discover all upnp devices, not only devices with AVTransport ability')
      print(' --play <url> - set current url for play and start playback it. In case of url is empty - continue playing recent media.')
      print(' --pause - pause current playback')
      print(' --stop - stop current playback')
      print(' --mute - mute playback')
      print(' --unmute - unmute playback')
      print(' --volume <vol> - set current volume for playback')
      print(' --seek <position in HH:MM:SS> - set current position for playback')
      print(' --timeout <seconds> - discover timeout')
      print(' --ssdp-version <version> - discover devices by protocol version, default 1')
      print(' --proxy - use local proxy on proxy port')
      print(' --proxy-port <port number> - proxy port to listen incomming connections from devices, default 8000')
      print(' --help - this help')

   def version():
      print(__version__)

   try:
      opts, args = getopt.getopt(sys.argv[1:], "hvd:t:i:", [   # information arguments
                                                               'help',
                                                               'version',
                                                               'log=',

                                                               # device arguments
                                                               'device=',
                                                               'ip=',

                                                               # action arguments
                                                               'play=',
                                                               'pause',
                                                               'stop',
                                                               'volume=',
                                                               'mute',
                                                               'unmute',
                                                               'seek=',


                                                               # discover arguments
                                                               'list',
                                                               'all',
                                                               'timeout=',
                                                               'ssdp-version=',

                                                               # transport info
                                                               'info',
                                                               'media-info',

                                                               # download proxy
                                                               'proxy',
                                                               'proxy-port='])
   except getopt.GetoptError:
      usage()
      sys.exit(1)

   device = ''
   url = ''
   vol = 10
   position = '00:00:00'
   timeout = 1
   action = ''
   logLevel = logging.WARN
   compatibleOnly = True
   ip = ''
   proxy = False
   proxy_port = 8000
   ssdp_version = 1
   for opt, arg in opts:
      if opt in ('-h', '--help'):
         usage()
         sys.exit(0)
      elif opt in ('-v', '--version'):
         version()
         sys.exit(0)
      elif opt in ('--log'):
         if arg.lower() == 'debug':
             logLevel = logging.DEBUG
         elif arg.lower() == 'info':
             logLevel = logging.INFO
         elif arg.lower() == 'warn':
             logLevel = logging.WARN
      elif opt in ('--all'):
         compatibleOnly = False
      elif opt in ('-d', '--device'):
         device = arg
      elif opt in ('-t', '--timeout'):
         timeout = float(arg)
      elif opt in ('--ssdp-version'):
         ssdp_version = int(arg)
      elif opt in ('-i', '--ip'):
         ip = arg
         compatibleOnly = False
         timeout = 10
      elif opt in ('--list'):
         action = 'list'
      elif opt in ('--play'):
         action = 'play'
         url = arg
      elif opt in ('--pause'):
         action = 'pause'
      elif opt in ('--stop'):
         action = 'stop'
      elif opt in ('--volume'):
         action = 'volume'
         vol = arg
      elif opt in ('--seek'):
         action = 'seek'
         position = arg
      elif opt in ('--mute'):
         action = 'mute'
      elif opt in ('--unmute'):
         action = 'unmute'
      elif opt in ('--info'):
         action = 'info'
      elif opt in ('--media-info'):
         action = 'media-info'
      elif opt in ('--proxy'):
         proxy = True
      elif opt in ('--proxy-port'):
         proxy_port = int(arg)

   logging.basicConfig(level=logLevel)

   st = URN_AVTransport_Fmt if compatibleOnly else SSDP_ALL
   allDevices = discover(name=device, ip=ip, timeout=timeout, st=st, ssdp_version=ssdp_version)
   if not allDevices:
      print('No compatible devices found.')
      sys.exit(1)

   if action in ('', 'list'):
      print('Discovered devices:')
      for d in allDevices:
         print(' {} {}'.format('[a]' if d.has_av_transport else '[x]', d))
      sys.exit(0)

   d = allDevices[0]
   print(d)

   if url.lower().replace('https://', '').replace('www.', '').startswith('youtube.'):
      import subprocess
      process = subprocess.Popen(['youtube-dl', '-g', url], stdout = subprocess.PIPE)
      url, err = process.communicate()

   if url.lower().startswith('https://'):
      proxy = True

   if proxy:
      ip = _get_serve_ip(d.ip)
      t = threading.Thread(target=runProxy, kwargs={'ip' : ip, 'port' : proxy_port})
      t.daemon = True
      t.start()
      time.sleep(2)

   if action == 'play':
      try:
         d.stop()
         url = 'http://{}:{}/{}'.format(ip, proxy_port, url) if proxy else url
         d.set_current_media(url=url)
         d.play()
      except Exception as e:
         print('Device is unable to play media.')
         logging.warn('Play exception:\n{}'.format(traceback.format_exc()))
         sys.exit(1)
   elif action == 'pause':
      d.pause()
   elif action == 'stop':
      d.stop()
   elif action == 'volume':
      d.volume(vol)
   elif action == 'seek':
      d.seek(position)
   elif action == 'mute':
      d.mute()
   elif action == 'unmute':
      d.unmute()
   elif action == 'info':
      print(d.info())
   elif action == 'media-info':
      print(d.media_info())

   if proxy:
      while running:
         time.sleep(30)



import os
from typing import List

import aria2p
import requests
from aria2p import Options

from aliyunpan.api.core import AliyunPan
from aliyunpan.api.models import *
from aliyunpan.api.req import *
from aliyunpan.api.utils import *
from aliyunpan.cli.config import Config
from aliyunpan.cli.tui import AliyunpanTUI
from aliyunpan.common import *
from aliyunpan.exceptions import InvalidRefreshToken, InvalidPassword, InvalidConfiguration, \
    ConfigurationFileNotFoundError, AliyunpanCode

__all__ = ['Commander']


class Commander:
    def __init__(self, init=True, *args, **kwargs):
        self._disk = AliyunPan()
        self._path_list = PathList(self._disk)
        self._req = Req()
        self._config = Config()
        self._task_config = Config(ROOT_DIR / Path('tasks.yaml'))
        self._share_link = 'aliyunpan://'
        self._print = Printer()
        self._host_url = 'https://www.aliyundrive.com/'
        self._aria2 = None
        self._config_set = {'~/.config/aliyunpan.yaml', '.config/aliyunpan.yaml', '~/aliyunpan.yaml', 'aliyunpan.yaml',
                            os.environ.get('ALIYUNPAN_CONF', '')}
        GLOBAL_VAR.tasks = self._task_config.read()
        GLOBAL_VAR.txt = ''
        if init:
            self.init(*args, **kwargs)

    def __del__(self):
        self._task_config.write(GLOBAL_VAR.tasks)
        if self._disk.refresh_token:
            try:
                self._config.update('refresh_token', self._disk.refresh_token)
            except AttributeError:
                pass

    def init(self, config_file=None, refresh_token=None, username=None, password=None, depth=3, timeout=None,
             drive_id=None, album=False):
        self._path_list.depth = depth
        self._req.timeout = timeout
        self._disk.drive_id = drive_id
        self._disk.album = album
        config_file_list = list(
            filter(lambda x: get_real_path(x).is_file(), map(lambda x: get_real_path(x), self._config_set)))
        if config_file:
            if not get_real_path(config_file).is_file():
                raise ConfigurationFileNotFoundError
            self._config.config_file = get_real_path(config_file)
        elif config_file_list and config_file_list[0]:
            self._config.config_file = config_file_list[0]
        if self._config.config_file and self._config.get('aria2'):
            aria2 = self._config.get('aria2')
        else:
            aria2 = {'host': 'http://localhost', 'port': 6800}
        if aria2:
            self._aria2 = self.aria2_init(**aria2)
        if refresh_token:
            if not len(refresh_token) == 32:
                raise InvalidRefreshToken
            self._disk.refresh_token = refresh_token
        elif username:
            if not password:
                raise InvalidPassword
            self._disk.login(username, password)
        elif config_file_list:
            refresh_token = self._config.get('refresh_token')
            username = self._config.get('username')
            password = self._config.get('password')
            if refresh_token:
                if not len(refresh_token) == 32:
                    raise InvalidRefreshToken
                self._disk.refresh_token = refresh_token
            elif username:
                if not password:
                    raise InvalidPassword
                self._disk.login(username, password)
            else:
                raise InvalidConfiguration
        else:
            raise ConfigurationFileNotFoundError

    def aria2_init(self, **kwargs):
        kwargs.setdefault('host', 'http://localhost')
        kwargs.setdefault('port', 6800)
        kwargs.setdefault('secret', '')
        self._aria2 = aria2p.API(
            aria2p.Client(
                **kwargs
            )
        )
        return self._aria2

    def ls(self, path='root', l=False, query=None):
        if query:
            file_info_list = self._path_list.get_file_info(self._disk.search(query))
        else:
            file_info_list = self._path_list.get_path_list(path, update=False)
        for i in file_info_list:
            if l:
                if i.type:
                    print(str_of_size(i.size), time.strftime('%d %b %H:%M', i.ctime), i.id, i.name)
                else:
                    print('-', time.strftime('%d %b %H:%M', i.ctime), i.id, i.name)
            else:
                print(i.name, end='\t')

    def get_path_list(self, path='root') -> List[FileInfo]:
        return self._path_list.get_path_list(path, update=False)

    def get_fid_list(self, file_id='root') -> List[FileInfo]:
        return self._path_list.get_fid_list(file_id, update=False)

    def tree(self, path='root', stdout=sys.stdout):
        return self._path_list.tree(path, stdout)

    def rm(self, path, file_id=None):
        if not file_id:
            file_id = self._path_list.get_path_fid(path, update=False)
        if file_id:
            file_id_ = self._disk.delete_file(file_id)
            if file_id_ == file_id:
                file_id = file_id_
                self._print.remove_info(path or file_id, status=False)
                self._path_list._tree.remove_node(file_id)
                self._print.print_line()
            else:
                file_id = False
        return file_id

    def rename(self, path, name):
        file_id = self._path_list.get_path_fid(path, update=False)
        status = False
        existed = False
        if file_id:
            file_id_ = self._disk.update_file(file_id, name)
            if file_id_ == AliyunpanCode.existed:
                file_id = False
                existed = True
            elif file_id_ == file_id:
                file_id = file_id_
                self._path_list.update_path_list(file_id=self._path_list._tree.get_node(file_id).data.pid, depth=0)
                status = True
        else:
            status = False
        self._print.rename_info(path, name=name, status=status, existed=existed)
        return file_id

    def mv(self, path, target_path):
        path = AliyunpanPath(path)
        target_path = AliyunpanPath(target_path)
        file_id = self._path_list.get_path_fid(path, update=False)
        target_file_id = self._path_list.get_path_fid(target_path, update=False)
        target_file_node = self._path_list._tree.get_node(target_file_id)
        if target_file_id and target_file_node:
            if target_file_node.data.type:
                raise FileExistsError
            else:
                _ = self._disk.move_file(file_id, target_file_id)
                if _ and file_id:
                    self._print.move_info(path, target_path, status=True)
                    self._path_list._tree.remove_node(file_id)
                    self._path_list.update_path_list(Path(target_path) / path, is_fid=False)
                else:
                    self._print.move_info(path, target_path, status=False)
                self._print.print_line()
                return _
        elif path.parent == target_path.parent:
            self.rename(path, target_path.name)

    def mkdir(self, path, name=None, parent_file_id=None):
        file_id_list = []
        if not parent_file_id or not name:
            path = AliyunpanPath(path)
            name = path.name
            if str(path) == 'root':
                return file_id_list
            file_id = self._path_list.get_path_fid(path, update=False)
            if file_id and file_id != 'root':
                return file_id_list
            parent_file_id = self._path_list.get_path_fid(path.parent, update=False)
            if not parent_file_id:
                file_id_list.extend(self.mkdir(path.parent))
                parent_file_id, _ = file_id_list[-1]
        r = self._disk.create_file(name, parent_file_id)
        try:
            file_id = r.json()['file_id']
        except KeyError:
            logger.debug(r.json()['message'])
            return False
        if file_id:
            self._print.mkdir_info(path, status=True)
            self._print.print_line()
            self._path_list._tree.create_node(tag=name, identifier=file_id, parent=parent_file_id,
                                              data=FileInfo(name=name, type=False, id=file_id, pid=parent_file_id))
            file_id_list.append((file_id, path))
        return file_id_list

    def upload(self, path, upload_path='root', timeout=10.0, retry=3, force=False, share=False, chunk_size=None,
               c=False):
        if isinstance(path, (str, AliyunpanPath, Path)):
            path_list = (path,)
        else:
            path_list = path
        result_list = []
        for path in path_list:
            if path:
                if self._share_link in str(path):
                    share_list = []
                    if share:
                        share_info = parse_share_url(path, self._disk.access_token)
                        file = self._path_list.get_path_node(share_info.name, update=False)
                        if file and not file.data.type:
                            path = path.replace(share_info.name, share_info.name + str(int(time.time())))
                            share_info = parse_share_url(path, self._disk.access_token)
                        if not self._path_list.get_path_fid(share_info.name, update=False):
                            self.upload_share(share_info)
                            self._path_list.update_path_list(depth=0)
                        if str(share_info.path) == 'root':
                            path_ = share_info.name
                        else:
                            path_ = share_info.path / share_info.name
                        for line in self.cat(path_).split('\n'):
                            if line.startswith(self._share_link):
                                share_list.append(parse_share_url(line, self._disk.access_token))
                        self.rm(path_)
                        if str(upload_path) == 'root':
                            upload_path = share_info.path
                        else:
                            upload_path /= share_info.path
                    else:
                        share_list = parse_share_url(path, self._disk.access_token)
                    return self.upload_share(share_list, upload_path, force)
                path = Path(path)
                if path.is_file():
                    if share:
                        share_list = []
                        with open(path, 'r', encoding='utf-8') as f:
                            while True:
                                line = f.readline()
                                if not line:
                                    break
                                if line.startswith(self._share_link):
                                    share_list.append(parse_share_url(line, self._disk.access_token))
                        return self.upload_share(share_list, upload_path, force)
                    else:
                        parent_file_id = self._path_list.get_path_fid(upload_path, update=False)
                        try:
                            result = self._disk.upload_file(
                                parent_file_id=parent_file_id, path=str(path),
                                upload_timeout=timeout, retry_num=retry, force=force, chunk_size=chunk_size, c=c)
                        except KeyboardInterrupt:
                            self.__del__()
                            raise
                        if result:
                            if isinstance(result, str):
                                file_id = result
                            else:
                                file_info = self._path_list.get_file_info(result)[0]
                                file_id = file_info.id
                                self._path_list._tree.create_node(tag=file_info.name, identifier=file_info.id,
                                                                  parent=parent_file_id, data=file_info)
                            result_list.append(file_id)
                elif path.is_dir():
                    if upload_path == 'root':
                        upload_path = '/'
                    upload_path = Path(upload_path)
                    upload_file_list = self.upload_dir(path, upload_path)
                    for file in upload_file_list:
                        try:
                            parent_file_id = self._path_list.get_path_fid(file[0], update=False)
                            result = self._disk.upload_file(
                                parent_file_id=parent_file_id, path=file[1],
                                upload_timeout=timeout, retry_num=retry, force=force, chunk_size=chunk_size, c=c)
                        except KeyboardInterrupt:
                            self.__del__()
                            raise
                        if result:
                            if isinstance(result, str):
                                file_id = result
                            else:
                                file_info = self._path_list.get_file_info(result)[0]
                                file_id = file_info.id
                                self._path_list._tree.create_node(tag=file_info.name, identifier=file_info.id,
                                                                  parent=parent_file_id, data=file_info)
                            result_list.append(file_id)
                else:
                    raise FileNotFoundError
                for file_hash, path in GLOBAL_VAR.file_set:
                    if file_hash in GLOBAL_VAR.tasks and GLOBAL_VAR.tasks[file_hash].upload_time:
                        if isinstance(GLOBAL_VAR.tasks[file_hash].path, str):
                            del GLOBAL_VAR.tasks[file_hash]
                        else:
                            try:
                                GLOBAL_VAR.tasks[file_hash].path.remove(path)
                            except ValueError:
                                pass
                            if not GLOBAL_VAR.tasks[file_hash].path:
                                del GLOBAL_VAR.tasks[file_hash]
        return result_list

    def upload_dir(self, path, upload_path):
        upload_path = upload_path / path.name
        if not self._path_list.get_path_fid(upload_path, update=False):
            self.mkdir(upload_path)
            self._print.print_line()
        upload_file_list = []
        for file in path.iterdir():
            if file.is_dir():
                upload_file_list.extend(self.upload_dir(file, upload_path))
            else:
                upload_file_list.append([upload_path, file])
        return upload_file_list

    def upload_share(self, share_info_list: List[ShareInfo], upload_path='root', force=False):
        if not isinstance(share_info_list, list):
            share_info_list = [share_info_list]
        if upload_path == 'root':
            upload_path = ''
        upload_path = AliyunpanPath(upload_path)
        folder_list = []
        file_list = []
        file_id_list = None
        for share_info in share_info_list:
            path = share_info.path
            if str(upload_path) in ('', '.') and str(path) == 'root':
                path = Path('')
            if str(upload_path) == 'root':
                upload_path = AliyunpanPath()
            p = upload_path / path
            if str(p) not in ('', '.'):
                file_id_list = self.mkdir(upload_path / path)
            if file_id_list:
                for file_id, path in file_id_list:
                    folder_list.append((file_id, upload_path / path))
        folder_list = list(set(folder_list))
        for share_info in share_info_list:
            share_info: ShareInfo
            path = share_info.path
            if str(upload_path) in ('', '.') and str(path) == 'root':
                path = Path()
            parent_file_id = self._path_list.get_path_fid(upload_path / path)
            result = self._disk.save_share_link(share_info.name, share_info.content_hash, share_info.proof_code,
                                                share_info.content_hash_name, share_info.size, parent_file_id, force)
            p = AliyunpanPath(upload_path / path / share_info.name)
            file_list.append((result, p))
            self._print.print_line()
            if result:
                self._print.upload_info(p, status=True, rapid_upload=True)
            else:
                self._print.upload_info(p, status=False)
        return folder_list, file_list

    def download(self, path, save_path=None, single_file=False, share=False, chunk_size=1048576, aria2=False, **kwargs):
        if not save_path:
            save_path = Path().cwd()
        save_path = Path(save_path)
        if isinstance(path, str):
            path_list = (path,)
        else:
            path_list = path
        kwargs.setdefault('referer', self._host_url)
        for path in path_list:
            if str(path).startswith(self._share_link) or share:
                folder_list, file_list = self.upload(path, share=share)
                folder_list = sorted(folder_list, key=lambda x: x[1])
                for file_id, path in folder_list:
                    p = save_path / path
                    try:
                        if not aria2:
                            p.mkdir(parents=True)
                            self._print.mkdir_info(p, status=True)
                    except FileExistsError:
                        pass
                for file_id, path in file_list:
                    if aria2:
                        kwargs.update({'dir': str((save_path / path).parent.absolute()), 'out': path.name})
                        self._aria2.add_uris([self._disk.get_download_url(file_id)], Options(self._aria2, kwargs))
                    else:
                        self.download_file(save_path / path, self._disk.get_download_url(file_id), chunk_size)
                for file_id, path in file_list:
                    self._path_list.update_path_list(path.parent, depth=0, is_fid=False)
                    try:
                        self.rm(path)
                    except FileNotFoundError:
                        pass
                for file_id, path in folder_list:
                    try:
                        self.rm(path)
                    except FileNotFoundError:
                        pass
                continue
            if isinstance(path, (Path, PurePosixPath, AliyunpanPath, str)):
                path = AliyunpanPath(path)
                node = self._path_list.get_path_node(path, update=False)
                if not node:
                    raise FileNotFoundError(path)
                file_node = node.data
                self._path_list.update_path_list(file_node.id)
                if file_node.type:
                    single_file = True
            else:
                file_node, path = path, path.name
            p = save_path / path
            if file_node.type:
                if single_file:
                    p = save_path / p.name
                if aria2:
                    kwargs.update({'dir': str(p.parent.absolute()), 'out': p.name})
                    self._aria2.add_uris([self._disk.get_download_url(file_node.id)], Options(self._aria2, kwargs))
                    self._print.download_info(p, status=True, aria2=True)
                else:
                    self._print.download_info(p)
                    self._print.print_line()
                    self.download_file(p, file_node.download_url, chunk_size)
                self._print.print_line()
            else:
                self.download(self._path_list.get_fid_list(file_node.id), save_path=save_path / p.name,
                              chunk_size=chunk_size, aria2=aria2, **kwargs)

    def download_file(self, path, url, chunk_size=1048576):
        try:
            path.parent.mkdir(parents=True)
            self._print.print_line()
            self._print.mkdir_info(path.parent, status=True)
            self._print.print_line()
        except FileExistsError:
            pass
        if path.exists():
            temp_size = path.stat().st_size
        else:
            temp_size = 0
        headers = {'Range': 'bytes=%d-' % temp_size}
        try:
            r = self._req.get(url, headers=headers, stream=True)
            file_size = int(r.headers['Content-Length'])
            if temp_size == file_size and file_size != 0:
                self._print.print_line()
                self._print.download_info(path, status=True)
                return True
            elif temp_size > file_size:
                mode = 'wb'
                temp_size = 0
            else:
                mode = 'ab'
            self._print.print_line()
            download_bar = DownloadBar(size=file_size)
            download_bar.update(refresh_line=False)
            with path.open(mode) as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    k = temp_size / file_size
                    download_bar.update(ratio=k, refresh_line=True)
                    if chunk:
                        temp_size += len(chunk)
                        f.write(chunk)
        except requests.exceptions.RequestException:
            self._print.refresh_line()
            self._print.download_info(path, status=False)
            self._print.print_line()
            return False
        self._print.download_info(path, status=True, t=download_bar.time, average_speed=download_bar.average_speed,
                                  refresh_line=True)
        self._print.print_line()
        return True

    def cat(self, path, encoding='utf-8'):
        file_node = self._path_list.get_path_node(path, update=False)
        if not file_node:
            raise FileNotFoundError(path)
        file = file_node.data
        self._path_list.update_path_list(file.id)
        r = self._req.get(file.download_url)
        r.encoding = encoding
        return r.text

    def share(self, path, expire_sec, share_link, download_link, save):
        def share_(path, file_id, parent_file=''):
            if path:
                file_node = self._path_list.get_path_node(path, update=False)
                if not file_node:
                    raise FileNotFoundError(path)
                file = file_node.data
                self._path_list.update_path_list(file.id)
            else:
                file = self._path_list._tree.get_node(file_id).data
            if file.type:
                share_txt = file.name.center(50, '-') + '\n'
                url = self._disk.get_download_url(file.id, expire_sec, file.category)
                if download_link:
                    share_txt += '下载链接'.center(50, '*') + '\n'
                    share_txt += url + '\n\n'
                if share_link:
                    share_txt += '分享链接'.center(50, '*') + '\n'
                    url_base64 = base64.b64encode(url.encode()).decode()
                    url = f'{self._share_link}{file.name}|{file.content_hash}|{url_base64}|{file.size}|{parent_file or "root"}'
                    share_txt += url + '\n'
                    share_txt += '导入链接'.center(50, '*') + '\n'
                    share_txt += f'python main.py upload "{url}"' + '\n\n'
                print(share_txt)
                GLOBAL_VAR.txt += share_txt
            else:
                for i in self._path_list.get_fid_list(file.id):
                    share_(path=None, file_id=i.id, parent_file=Path(parent_file) / file.name)

        GLOBAL_VAR.txt += '*' * 50 + '\n'
        GLOBAL_VAR.txt += '项目地址: https://github.com/wxy1343/aliyunpan' + '\n'
        GLOBAL_VAR.txt += '*' * 50 + '\n\n'
        if expire_sec is None:
            expire_sec = 14400
        share_(path, file_id=None)
        if save:
            file_name = Path(path).name + f'{int(time.time())}.txt'
            with open(file_name, 'w', encoding='utf-8') as f:
                f.write(GLOBAL_VAR.txt)
            print('文件导入'.center(50, '*'))
            print(f'python main.py upload -s {file_name}')
            print('链接导入'.center(50, '*'))
            file_id = self.upload(file_name)[0]
            print()
            if file_id:
                self._path_list.update_path_list(depth=0)
                file = self._path_list._tree.get_node(file_id).data
                url_base64 = base64.b64encode(self.disk.get_download_url(file_id).encode()).decode()
                url = f'{self._share_link}{Path(path).name}|{file.content_hash}|{url_base64}|{file.size}|root'
                print(f'python main.py upload -s "{url}"')

    def tui(self):
        aliyunpan_tui = AliyunpanTUI(self)
        aliyunpan_tui.run()

    def sync(self, path, upload_path, sync_time, time_out, chunk_size, retry, first=True):
        if first and path == 'root':
            self._print.print_info(
                'Do you really want to synchronize the root? This operation may delete all your files.', error=True)
            input('\nEnter to continue.')
        first = False
        path = AliyunpanPath(path)
        relative_path = AliyunpanPath(path.name)
        if str(relative_path) == '.':
            return self.sync(path.absolute(), upload_path, sync_time, time_out, chunk_size, retry, first=first)
        upload_path = AliyunpanPath(upload_path)
        p = upload_path / relative_path
        self._path_list.update_path_list(p, is_fid=False)
        file_id = self._path_list.get_path_fid(p, update=False)
        if not file_id:
            self.upload(path, upload_path, timeout=time_out, chunk_size=chunk_size, retry=retry)
            self._path_list.update_path_list(p, is_fid=False)
            file_id = self._path_list.get_path_fid(p, update=False)
        path_ = self._path_list._tree.to_dict(file_id, with_data=True)[str(relative_path)]
        change_file_list = self._path_list.check_path_diff(path, path_['children'] if 'children' in path_ else [])
        for path_ in change_file_list:
            relative_path = path.name / (path - path_)
            if path_.exists():
                self.upload(path_, upload_path / relative_path.parent, force=True, timeout=time_out,
                            chunk_size=chunk_size, retry=retry)
            else:
                self.rm(upload_path / relative_path)
        if sync_time:
            self._print.wait_info('等待{time}秒后再次同步', t=sync_time, refresh_line=True)
            self._print.refresh_line()
            self.sync(path, upload_path, sync_time, time_out, chunk_size, retry, first=first)

    def share_link(self, path_list, file_id_list=None, expiration=None):
        t = '' if expiration is None else time.time() + expiration
        if not file_id_list:
            file_id_list = [self._path_list.get_path_fid(path, update=False) for path in path_list if path]
        file_id_list = list(filter(None, file_id_list))
        if file_id_list:
            print(self._disk.share_link(file_id_list, t))
        else:
            raise FileNotFoundError

    def auto_refresh_token(self, refresh_time):
        print('Start to refresh token automatically.')
        while True:
            time.sleep(refresh_time)
            self._disk.token_refresh()
            print(time.strftime("%Y-%m-%d %H:%M:%S: ", time.localtime()) + self.disk.refresh_token)

    @property
    def req(self):
        return self._req

    @property
    def config(self):
        return self._config

    @property
    def path_list(self):
        return self._path_list

    @property
    def disk(self):
        return self._disk



import pathlib

from setuptools import setup

from aliyunpan.about import __version__

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')
install_requires = []
with open('requirements.txt', 'r') as f:
    while True:
        req = f.readline()
        if not req:
            break
        install_requires.append(req.strip('\n'))
setup(
    name='aliyunpan',
    version=__version__,
    description='aliyunpan cli',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/wxy1343/aliyunpan',
    author='wxy1343',
    author_email='1343890272@qq.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Framework :: Flask',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Video',
        'Topic :: Communications :: File Sharing',
        'Topic :: Utilities',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='aliyunpan, aliyundrive, cli, tui',
    py_modules=["main"],
    packages=['aliyunpan', 'aliyunpan/api', 'aliyunpan/cli', 'dlnap/dlnap'],
    python_requires='>=3.6, <4',
    install_requires=install_requires,
    data_files=[
        'requirements.txt'
    ],
    entry_points={
        'console_scripts': [
            'aliyunpan-cli=main:main',
            'aliyunpan=main:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/wxy1343/aliyunpan/issues',
        'Funding': 'https://donate.pypi.org',
        'Say Thanks!': 'https://saythanks.io/to/1343890272@qq.com',
        'Source': 'https://github.com/wxy1343/aliyunpan',
    },
    zip_safe=False
)



import base64
import ctypes
import hashlib
import inspect
import json
import logging
import os
import socket
import sys
from pathlib import Path

import requests
import rsa

__all__ = ['ROOT_DIR', 'logger', 'log_file', 'get_sha1', 'str_of_size', 'Iter', 'encrypt', 'parse_biz_ext',
           'stop_thread', 'get_open_port', 'get_real_path', 'get_proof_code', 'get_url_byte', 'get_file_byte']

ROOT_DIR = str(Path(os.environ.get('ALIYUNPAN_ROOT')).resolve().absolute()) if os.environ.get(
    'ALIYUNPAN_ROOT') else os.path.dirname(os.path.realpath(sys.argv[0]))
LOG_LEVEL = logging.INFO
log_file = ROOT_DIR + os.sep + 'aliyunpan.log'
fmt_str = "%(asctime)s [%(filename)s:%(lineno)d] %(funcName)s %(levelname)s - %(message)s"
logging.basicConfig(level=LOG_LEVEL,
                    format=fmt_str,
                    stream=open(log_file, 'a', encoding='utf-8'),
                    datefmt="%Y-%m-%d %H:%M:%S")
logger = logging.getLogger('aliyunpan')
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


def get_sha1(path, split_size=524288):
    logger.info(f'Calculate sha1 of file {path}.')
    file_size = os.path.getsize(path)
    from aliyunpan.common import HashBar
    hash_bar = HashBar(size=file_size)
    hash_bar.hash_info(path, size=file_size)
    hash_bar.print_line()
    hash_bar.update(refresh_line=False)
    with open(path, 'rb') as f:
        sha1 = hashlib.sha1()
        count = 0
        while True:
            chunk = f.read(split_size)
            k = ((count * split_size) + len(chunk)) / file_size if file_size else 0
            hash_bar.update(ratio=k, refresh_line=True)
            if not chunk:
                break
            count += 1
            sha1.update(chunk)
        content_hash = sha1.hexdigest()
    logger.info(f'The SHA1 of file {path} is {content_hash}.')
    hash_bar.refresh_line()
    hash_bar.hash_info(path, status=True, size=file_size, refresh_line=True)
    hash_bar.print_line()
    return content_hash


def get_proof_code(bys: bytes) -> str:
    proof_code = base64.b64encode(bys).decode()
    return proof_code


def get_file_byte(path: Path, access_token: str = None):
    n1 = int(hashlib.md5(access_token.encode()).hexdigest()[:16], 16)
    n2 = path.stat().st_size
    n3 = (n1 % n2) if n2 else 0
    with path.open('rb') as f:
        f.seek(n3)
        return f.read(min(n3 + 8, n2) - n3)


def get_url_byte(url: str, access_token: str = None, file_size: int = None):
    from aliyunpan.api.req import Req
    req = Req()
    if not file_size:
        file_size = int(req.get(url, stream=True).headers.get('Content-Length'))
    n1 = int(hashlib.md5(access_token.encode()).hexdigest()[:16], 16)
    n2 = file_size
    n3 = (n1 % n2) if n2 else 0
    headers = {'Range': f'bytes={n3}-{min(n3 + 8, n2) - 1}'}
    r: requests.Response = req.get(url, headers=headers, stream=True)
    return r.content


def str_of_size(size, decimal=3, tuple_=False):
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    integer = int(size)
    level = 0
    while integer >= 1024:
        integer //= 1024
        level += 1
    if level + 1 > len(units):
        level = -1
    size = round(size / (1024 ** level), decimal) if decimal else size
    if tuple_:
        return size, units[level]
    return f'{size}{units[level]}'


class Iter:

    def __init__(self, IterObj):
        self.iter = IterObj

    def __getitem__(self, index):
        return self.iter[index]

    def __len__(self):
        return len(self.iter)


# RSA encrypt
PUBLIC_KEY = b'-----BEGIN PUBLIC KEY-----\n' \
             b'MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDTvO8fAEJPMmHIkyP6jN+hK7rE\n' \
             b'ANn+i7Yn6NJ6RL1dWdzlWRNdZ4qBQ761uNcFbE4ficTh8VJHBiW3tBlEqX8C2m9g\n' \
             b'WkmpPsbrnLry56wrJqNUzmnrJllT0sKeOV1tjBzbaIl4VRqg91IfKQA1+tOBF42g\n' \
             b'vqj55q3OOQIPUTEz+wIDAQAB\n' \
             b'-----END PUBLIC KEY-----'


def encrypt(password):
    MAP = "0123456789abcdefghijklmnopqrstuvwxyz"
    rsa_result = rsa.encrypt(
        password.encode(),
        rsa.PublicKey.load_pkcs1_openssl_pem(PUBLIC_KEY)
    )
    ans = ''
    for byte in rsa_result:
        ans += MAP[byte >> 4]
        ans += MAP[byte & 0x0F]
    return ans


def parse_biz_ext(biz_ext):
    biz_ext = base64.b64decode(biz_ext).decode('gbk')
    logger.debug(biz_ext)
    return json.loads(biz_ext)


def _async_raise(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    tid = ctypes.c_long(tid)
    if not inspect.isclass(exctype):
        exctype = type(exctype)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


def stop_thread(thread):
    _async_raise(thread.ident, SystemExit)


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def get_real_path(path):
    return Path(path).expanduser().resolve().absolute()
